@@ -20,6 +20,12 @@
 ;;; is free to :USE (PACKAGE-USE-LIST :CL-USER) or whatever.
 (defglobal *!default-package-use-list* nil)
 
+(defmacro sanitize-nicknames (name list)
+  `(let ((list ,list))
+     (when list
+       (remove ,name (remove-duplicates (stringify-string-designators list) :test 'string=)
+               :test 'string=))))
+
 (defun make-package (name &key
                           (use '#.*!default-package-use-list*)
                           nicknames
@@ -31,70 +37,52 @@ list. :INTERNAL-SYMBOLS and :EXTERNAL-SYMBOLS are estimates for the number of
 internal and external symbols which will ultimately be present in the package.
 The default value of USE is implementation-dependent, and in this
 implementation it is ~S." *!default-package-use-list*)
-  (prog ((name (stringify-string-designator name))
-         (nicks (stringify-string-designators nicknames))
+  (let* ((name (stringify-string-designator name))
+         (nicks (sanitize-nicknames name nicknames))
          (package
+          ;; a "resolved" package is not in the global name->package mapping yet,
+          ;; which is why the FIND-PACKAGE / CERROR below does not signal.
           (or (resolve-deferred-package name)
-              (resolve-rehoming-package name)
-              (%make-package (make-symbol-hashset internal-symbols)
-                             (make-symbol-hashset external-symbols))))
-         clobber)
-   :restart
-     (when (find-package name)
-       ;; ANSI specifies that this error is correctable.
-       (signal-package-cerror
-        name
-        "Clobber existing package."
-        "A package named ~S already exists" name)
-       (setf clobber t))
-     (with-package-graph ()
-         ;; Check for race, signal the error outside the lock.
-         (when (and (not clobber) (find-package name))
-           (go :restart))
-         (setf (package-%name package) name)
-         ;; Do a USE-PACKAGE for each thing in the USE list so that checking for
-         ;; conflicting exports among used packages is done.
-         (use-package use package)
-         ;; FIXME: ENTER-NEW-NICKNAMES can fail (ERROR) if nicknames are illegal,
-         ;; which would leave us with possibly-bad side effects from the earlier
-         ;; USE-PACKAGE (e.g. this package on the used-by lists of other packages,
-         ;; but not in *PACKAGE-NAMES*, and possibly import side effects too?).
-         ;; Perhaps this can be solved by just moving ENTER-NEW-NICKNAMES before
-         ;; USE-PACKAGE, but I need to check what kinds of errors can be caused by
-         ;; USE-PACKAGE, too.
-         (%enter-new-nicknames package nicks)
-         ;; The name table is actually multi-writer concurrent, but due to
-         ;; lazy removal of :DELETED entries we want to enforce a single-writer.
-         ;; We're inside WITH-PACKAGE-GRAPH so this is already synchronized with
-         ;; other MAKE-PACKAGE operations, but we need the additional lock
-         ;; so that it synchronizes with RENAME-PACKAGE.
-         (with-package-names (table)
-           (%register-package table name package))
-         (atomic-incf *package-names-cookie*)
-         (when (boundp 'sb-c::*compilation*)
-           (setf (sb-c::package-environment-changed sb-c::*compilation*) t))
-         (return package))
-     (bug "never")))
-
-(flet ((remove-names (package name-table keep-primary-name)
-         ;; An INFO-HASHTABLE does not support REMHASH. We can simulate it
-         ;; by changing the value to :DELETED.
-         ;; (NIL would be preferable, but INFO-GETHASH does not return
-         ;; a secondary value indicating whether the NIL was by default
-         ;; or found, not does it take a different default to return).
-         ;; At some point the table might contain more deleted values than
-         ;; useful values. We call %REBUILD-PACKAGE-NAMES to rectify that.
-         (dx-let ((names (cons (package-name package)
-                               (package-nicknames package)))
-                  (i 0))
-           (when keep-primary-name (pop names))
-           (dolist (name names)
-             ;; Aver that the following SETF doesn't insert a new <k,v> pair.
-             (aver (info-gethash name name-table))
-             (setf (info-gethash name name-table) :deleted)
-             (incf i))
-           (incf (info-env-tombstones name-table) i))
-         nil))
+              (%make-package (make-symbol-table internal-symbols)
+                             (make-symbol-table external-symbols))))
+         (existing-pkg)
+         (namelist (cons name nicks))
+         (conflict))
+    (setf (symtbl-package (package-external-symbols package)) package)
+    (with-package-names (table) ; get exclusive use of name -> package mapping
+      ;; If the loop runs to completion, then insert all names,
+      ;; which also assigns the %NAME and KEYS slots of the package.
+      (dolist (string namelist (package-registry-update package namelist))
+        (when (setq existing-pkg (%get-package string table))
+          (return (setq conflict string)))))
+    (when existing-pkg
+      ;; Of the possible ways to continue from the "package already exists" error,
+      ;; I've seen implementations offer these:
+      ;;  - pick a different name for the new package
+      ;;  - return the existing package as-is
+      ;; Other ways to proceed might be:
+      ;;  - rename the existing package to something different (mostly harmless)
+      ;;  - first delete the existing package (potentially nontrivial)
+      ;; SBCL had the weirdest of all solutions: alter the existing name->package mapping
+      ;; while leaving the old package with a name, but not findable by that name.
+      ;; Imho the safe assumption is that the user wants the same package back.
+      ;; All sorts of crazy restarts would in theory be possible, such as if one nickname
+      ;; that you specified (N1) already finds package P1, and another nickname (N2) finds P2
+      ;; and P1 and P2 are distinct. What are we supposed to do? Figure our that you meant
+      ;; to return the existing P1, but give it an additional nickname, and delete P2?
+      ;; Also bear in mind that for most purposes, a package's multiple "names" are
+      ;; equivalent; one is canonical for printing symbols homed in that package.
+      ;; Global nicknames are an absurdly unnecessary part of the language.
+      (signal-package-cerror name "Return the existing package."
+                             "A package named ~S already exists" conflict)
+      ;; We don't do the USE in this case.
+      ;; It's OK given the lack of guidance in CLHS about how to "continue".
+      (return-from make-package existing-pkg))
+    (atomic-incf *package-names-cookie*)
+    (when (boundp 'sb-c::*compilation*)
+      (setf (sb-c::package-environment-changed sb-c::*compilation*) t))
+    (use-package use package)
+    package))
 
 ;;; Change the name if we can, blast any old nicknames and then
 ;;; add in any new ones.
@@ -121,52 +109,50 @@ implementation it is ~S." *!default-package-use-list*)
 ;;; was given. I see no reason to be more strict than the spec would have it be.
 (defun rename-package (package-designator name &optional (nicknames ()))
   "Changes the name and nicknames for a package."
-  (prog ((nicks (stringify-string-designators nicknames)))
-   :restart
-     (let ((package (find-undeleted-package-or-lose package-designator))
-           ;; This is the "weirdness" alluded to. Do it in the loop in case
-           ;; the stringified value changes on restart when NAME is a package.
-           (name (stringify-package-designator name))
-           (found (find-package name)))
-       (unless (or (not found) (eq found package))
-         (signal-package-error name
-                               "A package named ~S already exists." name))
-       (with-single-package-locked-error ()
-         (unless (and (string= name (package-name package))
-                      (null (set-difference nicks (package-nicknames package)
-                                            :test #'string=)))
-           (assert-package-unlocked
-            package "renaming as ~A~@[ with nickname~*~P ~1@*~{~A~^, ~}~]"
-            name nicks (length nicks)))
-         (with-package-names (table)
-           ;; Check for race conditions now that we have the lock.
-           (unless (eq package (find-package package-designator))
-             (go :restart))
-           ;; Do the renaming.
-           ;; As a special case, do not allow the package to transiently disappear
-           ;; if PACKAGE-NAME is unchanged. This avoids glitches with build systems
-           ;; which try to operate on subcomponents in parallel, where one of the
-           ;; built subcomponents needs to add nicknames to an existing package.
-           ;; We could be clever here as well by not removing nicknames that
-           ;; will ultimately be re-added, but that didn't seem as critical.
-           (let ((keep-primary-name (string= name (package-%name package))))
-             (remove-names package table keep-primary-name)
-             (unless keep-primary-name
-               (%register-package table name package)))
-           (setf (package-%name package) name
-                 (package-%nicknames package) ()))
-         ;; Adding each nickname acquires and releases the table lock,
-         ;; because it's potentially interactive (on failure) and therefore
-         ;; not ideal to hold the lock for the entire duration.
-         (%enter-new-nicknames package nicks))
-       (atomic-incf *package-names-cookie*)
-       (when (boundp 'sb-c::*compilation*)
-         (setf (sb-c::package-environment-changed sb-c::*compilation*) t))
-       (return package))))
+  ;; CLHS says:
+  ;; "The consequences are undefined if new-name or any new-nickname
+  ;; conflicts with any existing package names."
+  ;; Signaling an error is what most implementations do. So shall we now.
+  ;; (There is no portable standard way to proceed)
+  (let* ((package (find-undeleted-package-or-lose package-designator))
+         ;; This potentially allows the "weirdness" alluded to above
+         (name (stringify-package-designator name))
+         (nicknames (sanitize-nicknames name nicknames))
+         (namelist (cons name nicknames))
+         (conflict))
+    (with-single-package-locked-error ()
+      (unless (and (string= name (package-name package))
+                   (null (set-difference nicknames (package-%nicknames package)
+                                         :test #'string=))
+                   (null (set-difference (package-%nicknames package) nicknames
+                                         :test #'string=)))
+        (assert-package-unlocked
+         package "renaming as ~A~@[ with nickname~*~P ~1@*~{~A~^, ~}~]"
+         name nicknames (length nicknames))))
+    (with-package-names (table)
+      ;; get exclusive use of name -> package mapping
+      ;; and also prevent concurrent modification to this package's names.
+      (dolist (string namelist (package-registry-update package namelist))
+        (let ((found (%get-package string table)))
+          (cond ((eq found package))
+                (found (return (setq conflict string)))))))
+    (cond (conflict
+           (signal-package-error
+            ;; avoid saying "another package ... has name X" so the pedants
+            ;; don't complain when X is a "nickname" rather than "name"
+            package "Another package is already accessible via name ~S" conflict))
+          (t
+           (atomic-incf *package-names-cookie*)
+           (when (boundp 'sb-c::*compilation*)
+             (setf (sb-c::package-environment-changed sb-c::*compilation*) t))
+           package))))
 
 (defun delete-package (package-designator)
   "Delete the package designated by PACKAGE-DESIGNATOR from the package
   system data structures."
+  (when (and (packagep package-designator)
+             (not (package-%name package-designator))) ; already deleted
+    (return-from delete-package nil))
   (tagbody :restart
      (let ((package (find-package package-designator)))
        (cond ((not package)
@@ -176,8 +162,6 @@ implementation it is ~S." *!default-package-use-list*)
                "Ignore."
                "There is no package named ~S." package-designator)
               (return-from delete-package nil))
-             ((not (package-name package)) ; already deleted
-              (return-from delete-package nil))
              (t
               (with-single-package-locked-error
                   (:package package "deleting package ~A" package)
@@ -203,37 +187,51 @@ implementation it is ~S." *!default-package-use-list*)
                   (dolist (used (package-use-list package))
                     (unuse-package used package))
                   (setf (package-%local-nicknames package) nil)
-                  (let ((rehoming-package (make-rehoming-package package)))
-                    (flet ((nullify-home (symbols)
-                             (dovector (x (symtbl-cells symbols))
-                               (when (and (symbolp x)
-                                          (eq (symbol-package x) package))
-                                 (%set-symbol-package x rehoming-package)))))
-                      (nullify-home (package-internal-symbols package))
-                      (nullify-home (package-external-symbols package))))
-                  (with-package-names (table)
+                  (flet ((nullify-home (symbols)
+                           (dovector (x (symtbl-cells symbols))
+                             (when (and (symbolp x)
+                                        (eq (symbol-package x) package))
+                               (%set-symbol-package x nil)))))
+                    (nullify-home (package-internal-symbols package))
+                    (nullify-home (package-external-symbols package)))
+                  (with-package-names ()
+                    (package-registry-update package nil)
                     (awhen (package-id package)
                       (setf (aref *id->package* it) nil (package-id package) nil))
-                    (remove-names package table nil)
                     (setf (package-%name package) nil
                           ;; Setting PACKAGE-%NAME to NIL is required in order to
                           ;; make PACKAGE-NAME return NIL for a deleted package as
                           ;; ANSI requires. Setting the other slots to NIL
-                          ;; and blowing away the SYMBOL-HASHSETs is just done
+                          ;; and blowing away the SYMBOL-TABLEs is just done
                           ;; for tidiness and to help the GC.
-                          (package-%nicknames package) nil))
+                          (package-keys package) #()))
                   (atomic-incf *package-names-cookie*)
                   (when (boundp 'sb-c::*compilation*)
                     (setf (sb-c::package-environment-changed sb-c::*compilation*) t))
-                  (setf (package-%use-list package) nil
-                        (package-tables package) #()
+                  (setf (package-tables package) #()
                         (package-%shadowing-symbols package) nil
-                        (package-internal-symbols package)
-                        (make-symbol-hashset 0)
-                        (package-external-symbols package)
-                        (make-symbol-hashset 0)))
+                        (package-internal-symbols package) (make-symbol-table 0)
+                        (package-external-symbols package) (make-symbol-table 0)))
                 (return-from delete-package t)))))))
-) ; end FLET
+
+;;; Possible FIXME:
+;;;   After doing these 2 things:
+;;;    (defpackage "P" (:nicknames "PNICK" "PNICK2"))
+;;;    (defpackage "P" (:nicknames "NN"))
+;;;   how many nicknames has "P" - 1, 2, 3?
+;;;   The case for 1 is that DEFPACKAGE says:
+;;;      "The arguments to :nicknames set the package's nicknames to the supplied names."
+;;       [ECL interprets the redefinition thusly]
+;;;   The case for 2 is that it says:
+;;;      "If defined-package-name already refers to an existing package,
+;;;       the name-to-package mapping for that name is not changed."
+;;;      [ABCL interprets the redefinition thusly]
+;;;   The case for 3 is that nicknames are cumulative.
+;;;      [SBCL interprets the redefinition thusly]
+;;; I suspect that any treatment is fine as long as its documented.
+;;; Relatedly, ECL and ABCL take a redefinition with 0 nicknames as follows:
+;;;  (defpackage "P" (:nicknames))
+;;; to imply _no_ _change_ to existing nicknames, while CLISP removes any nicknames.
 
 (defmacro defpackage (package &rest options)
   #.(format nil
@@ -421,7 +419,7 @@ implementation it is ~S." *!default-package-use-list*)
                        imports interns
                        exports implement local-nicknames
                        lock doc-string)
-  (%enter-new-nicknames package nicknames)
+  (rename-package package (package-name package) nicknames)
   ;; 1. :shadow and :shadowing-import-from
   ;;
   ;; shadows is a list of strings, shadowing-imports is a list of symbols.
@@ -649,7 +647,7 @@ specifies to signal a warning if SWANK package is in variance, and an error othe
   (let* ((string (truly-the simple-string
                             (stringify-string-designator string-designator)))
          (length (length string))
-         (hash (compute-symbol-hash string length))
+         (hash (calc-symbol-name-hash string length))
          (result (list nil)))
     (do-packages (p) ; FIXME: should not acquire package-names lock
       (add-to-bag-if-found (package-internal-symbols p) string length hash result)
@@ -703,7 +701,7 @@ specifies to signal a warning if SWANK package is in variance, and an error othe
                     #+sb-unicode
                     (= widetag sb-vm:simple-character-string-widetag))
                 (when (search string obj :test #'char-equal)
-                  (push (cons (compute-symbol-hash obj (length obj)) obj) candidates)))
+                  (push (cons (calc-symbol-name-hash obj (length obj)) obj) candidates)))
                (t
                 (return-from done))))
        :read-only))
@@ -750,12 +748,79 @@ specifies to signal a warning if SWANK package is in variance, and an error othe
     (briefly-describe-symbol symbol))
   (values))
 
+(defun perfectly-hash-package (designator)
+  (flet ((rehash (table &aux (cells (symtbl-%cells table)))
+           (when (functionp (car cells))
+             (return-from rehash)) ; already hashed
+           (let* ((cells (cdr cells))
+                  (hashes (map '(simple-array (unsigned-byte 32) (*))
+                               (lambda (symbol) (ldb (byte 32 0) (symbol-hash symbol)))
+                               (remove-if-not #'symbolp cells))))
+             (when (or (< (length (remove-duplicates hashes)) (length hashes))
+                       ;; limitation of perfecthash, also probably no benefit
+                       (< (length hashes) 5))
+               (return-from rehash))
+             (let* ((hash-expr (sb-c:make-perfect-hash-lambda hashes))
+                    (fun (compile nil hash-expr))
+                    (new-cells (make-array (length hashes) :initial-element 0)))
+               (dovector (s cells)
+                 (when (symbolp s)
+                   (let ((hash (funcall fun (ldb (byte 32 0) (symbol-hash s)))))
+                     (aver (eq (svref new-cells hash) 0))
+                     (setf (svref new-cells hash) s))))
+               (setf (symtbl-%cells table) (cons fun new-cells)
+                     (symtbl-size table) (length hashes)
+                     (symtbl-free table) 0
+                     (symtbl-deleted table) 0)))))
+    (let ((package (find-package designator)))
+      (rehash (package-internal-symbols package))
+      (rehash (package-external-symbols package)))))
+
+;;; Reorganize both hashsets of all packages, called by SAVE-LISP-AND-DIE.
+;;; With a few exceptions, tables are saved at 100% load factor and a perfect hash.
+;;; The danger of attempting to achieve such high load using an open-addressing table
+;;; is that despite the robinhood algorithm rearranging elements, there could always
+;;; be a tremendously bad probe sequence for some symbol, not to mention that unused
+;;; cells must be scattered throughout, to assure probing always terminates.
+;;; The relevant benchmark is how quickly we can return that a symbol is NOT found.
+;;; Test case:
+;;; * (defvar *l* (let (l) (do-all-symbols (s l) (push s l))))
+;;; * (defun f (list &aux (n 0))
+;;;    (dolist (s list n) (if (find-symbol (string s) #.(find-package "CL")) (incf n))))
+;;; * (time (f *l*))
+;;;
+;;; Freshly restarted image after save-lisp-and-die without perfect hash:
+;;;   0.017 seconds of real time
+;;;
+;;; Freshly restarted image with perfect hash:
+;;;   0.012 seconds of real time
+;;;
+(defun tune-hashset-sizes-of-all-packages ()
+  (flet ((tune-table-size (desired-lf table)
+           (resize-symbol-table table (%symtbl-count table) t desired-lf)
+           ;; The APROPOS-LIST R/O scan optimization is inadmissible if no R/O space
+           #-darwin-jit (setf (symtbl-modified table) nil)))
+    (dolist (package (list-all-packages))
+      ;; Choose load factor based on whether INTERN is expected at runtime
+      (let ((lf (cond ((eq (the package package) *keyword-package*) 60/100)
+                      ((eq (package-id package) +package-id-user+) 8/10)
+                      (t 1))))
+        (cond ((= lf 1)
+               (perfectly-hash-package package))
+              (t
+               (tune-table-size lf (package-internal-symbols (truly-the package package)))
+               (tune-table-size lf (package-external-symbols package))))))))
+
+;;; This function is mainly of interest to developers, should we remove it
+;;; from the image?
 (export 'show-package-utilization)
 (defun show-package-utilization (&aux (tot-ncells 0))
-  (flet ((symtbl-metrics (table &aux (cells (symtbl-%cells table))
-                                     (reciprocals (car cells))
-                                     (vec (cdr cells))
-                                     (nslots (length vec)))
+  (flet ((metrics (table &aux (cells (symtbl-%cells table))
+                              (reciprocals (car cells))
+                              (vec (cdr cells))
+                              (nslots (length vec)))
+           (when (functionp reciprocals)
+             (return-from metrics (values 1 1 1))) ; 1 probe max+avg, 100% load
            (flet ((probe-seq-len (symbol)
                     (let* ((name-hash (sxhash symbol))
                            (index (symbol-table-hash 1 name-hash nslots))
@@ -778,8 +843,8 @@ specifies to signal a warning if SWANK package is in variance, and an error othe
   (dolist (pkg (list-all-packages))
     (binding* ((ext (package-external-symbols pkg))
                (int (package-internal-symbols pkg))
-               ((ext-max ext-psl ext-lf) (symtbl-metrics ext))
-               ((int-max int-psl int-lf) (symtbl-metrics int))
+               ((ext-max ext-psl ext-lf) (metrics ext))
+               ((int-max int-psl int-lf) (metrics int))
                (ncells (+ (length (symtbl-cells int))
                           (length (symtbl-cells ext)))))
       (incf tot-ncells ncells)
@@ -789,3 +854,6 @@ specifies to signal a warning if SWANK package is in variance, and an error othe
                     (list int-max int-psl int-lf))
               (package-name pkg))))
   (format t "~8d~%" tot-ncells)))
+
+(let ((package (find-package "SB-SEQUENCE")))
+  (rename-package package package (list "SEQUENCE")))
