@@ -9,14 +9,10 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB-IMPL") ;(SB-IMPL, not SB!IMPL, since we're built in warm load.)
+(in-package "SB-IMPL")
 
 (defparameter *inspect-length* 10)
 
-;;; When *INSPECT-UNBOUND-OBJECT-MARKER* occurs in a parts list, it
-;;; indicates that that a slot is unbound.
-(defvar *inspect-unbound-object-marker* (gensym "INSPECT-UNBOUND-OBJECT-"))
-
 (defun inspector (object input-stream output-stream)
   (declare (ignore input-stream))
   (catch 'quit-inspect
@@ -34,7 +30,7 @@
 (defun inspect (object)
   (funcall *inspect-fun* object *standard-input* *standard-output*))
 
-(defvar *help-for-inspect*
+(defconstant-eqx +help-for-inspect+
   "
 help for INSPECT:
   Q, E        -  Quit the inspector.
@@ -46,7 +42,7 @@ help for INSPECT:
 Within the inspector, the special variable SB-EXT:*INSPECTED* is bound
 to the current inspected object, so that it can be referred to in
 evaluated expressions.
-")
+" #'equal)
 
 (defun %inspect (*inspected* s)
   (named-let redisplay () ; "LAMBDA, the ultimate GOTO":-|
@@ -74,7 +70,7 @@ evaluated expressions.
                (cond ((< -1 command elements-length)
                       (let* ((element (nth command elements))
                              (value (if named-p (cdr element) element)))
-                        (cond ((eq value *inspect-unbound-object-marker*)
+                        (cond ((eq value sb-pcl:+slot-unbound+)
                                (format s "~%That slot is unbound.~%")
                                (return-from %inspect (reread)))
                               (t
@@ -99,7 +95,7 @@ evaluated expressions.
                (:r
                 (return-from %inspect (redisplay)))
                ((:h :? :help)
-                (write-string *help-for-inspect* s)
+                (write-string +help-for-inspect+ s)
                 (return-from %inspect (reread)))
                (t
                 (eval-for-inspect command s)
@@ -120,7 +116,8 @@ evaluated expressions.
   (let ((*suppress-print-errors*
          (if (subtypep 'serious-condition *suppress-print-errors*)
              *suppress-print-errors*
-             'serious-condition)))
+             'serious-condition))
+        (unbound (load-time-value (make-unprintable-object "unbound slot") t)))
     (format stream "~%~A" description)
     (loop for element in elements
        for index from 0
@@ -128,10 +125,9 @@ evaluated expressions.
               (if named-p
                   (values (cdr element) (car element))
                   element)
-            (format stream "~W. ~@[~A: ~]~W~%"
-                    index name (if (eq value *inspect-unbound-object-marker*)
-                                   "unbound"
-                                   value))))))
+            (when (unbound-marker-p value)
+              (setf value unbound))
+            (format stream "~W. ~@[~A: ~]~W~%" index name value)))))
 
 ;;;; INSPECTED-PARTS
 
@@ -161,10 +157,10 @@ evaluated expressions.
                 (cons "Package" (symbol-package object))
                 (cons "Value" (if (boundp object)
                                   (symbol-value object)
-                                  *inspect-unbound-object-marker*))
+                                  sb-pcl:+slot-unbound+))
                 (cons "Function" (if (fboundp object)
                                      (symbol-function object)
-                                     *inspect-unbound-object-marker*))
+                                     sb-pcl:+slot-unbound+))
                 (cons "Plist" (symbol-plist object)))))
 
 (defun inspected-structure-elements (object)
@@ -172,9 +168,10 @@ evaluated expressions.
         (info (layout-info (sb-kernel:layout-of object))))
     (when (sb-kernel::defstruct-description-p info)
       (dolist (dd-slot (dd-slots info) (nreverse parts-list))
-        (push (cons (dsd-name dd-slot)
-                    (funcall (dsd-accessor-name dd-slot) object))
-              parts-list)))))
+        (let* ((reader (dsd-reader dd-slot (neq (dd-type info) 'structure)))
+               (index (dsd-index dd-slot))
+               (value (funcall reader object index)))
+          (push (cons (dsd-name dd-slot) value) parts-list))))))
 
 (defmethod inspected-parts ((object structure-object))
   (values (format nil "The object is a STRUCTURE-OBJECT of type ~S.~%"
@@ -182,6 +179,11 @@ evaluated expressions.
           t
           (inspected-structure-elements object)))
 
+(defmethod inspected-parts ((object pathname))
+  (values (format nil "The object is a ~S.~%" (type-of object))
+          t
+          (inspected-structure-elements object)))
+
 (defun inspected-standard-object-elements (object)
   (let ((reversed-elements nil)
         (class-slots (sb-pcl::class-slots (class-of object))))
@@ -189,7 +191,7 @@ evaluated expressions.
       (let* ((slot-name (slot-value class-slot 'sb-pcl::name))
              (slot-value (if (slot-boundp object slot-name)
                              (slot-value object slot-name)
-                             *inspect-unbound-object-marker*)))
+                             sb-pcl:+slot-unbound+)))
         (push (cons slot-name slot-value) reversed-elements)))))
 
 (defmethod inspected-parts ((object standard-object))
@@ -220,27 +222,13 @@ evaluated expressions.
           ;; to DESCRIBE from the inspector.
           (list*
            (cons "Lambda-list" (%fun-lambda-list object))
-           (cons "Ftype" (%fun-type object))
+           (cons "Ftype" (%fun-ftype object))
            (when (closurep object)
              (list
               (cons "Closed over values" (%closure-values object)))))))
 
-#+sb-eval
-(defmethod inspected-parts ((object sb-eval:interpreted-function))
-  (multiple-value-bind (defn closurep name) (function-lambda-expression object)
-    (declare (ignore closurep))
-    (values (format nil "The object is an interpreted function named ~S.~%" name)
-            t
-            ;; Defined-from stuff used to be here. Someone took
-            ;; it out. FIXME: We should make it easy to get
-            ;; to DESCRIBE from the inspector.
-            (list
-             (cons "Lambda-list" (sb-eval:interpreted-function-lambda-list object))
-             (cons "Definition" defn)
-             (cons "Documentation" (sb-eval:interpreted-function-documentation object))))))
-
-#+sb-fasteval
-(defmethod inspected-parts ((object sb-interpreter:interpreted-function))
+#+(or sb-eval sb-fasteval)
+(defmethod inspected-parts ((object interpreted-function))
   (multiple-value-bind (defn closurep name) (function-lambda-expression object)
     (declare (ignore closurep))
     (values (format nil "The object is an interpreted function named ~S.~%" name)
@@ -251,18 +239,19 @@ evaluated expressions.
             (list
              (cons "Lambda-list" (%fun-lambda-list object))
              (cons "Definition" defn)
-             (cons "Documentation" (%fun-doc object))))))
+             (cons "Documentation" (documentation object t))))))
 
 (defmethod inspected-parts ((object vector))
-  (values (format nil
-                  "The object is a ~:[~;displaced ~]VECTOR of length ~W.~%"
-                  (and (array-header-p object)
-                       (%array-displaced-p object))
-                  (length object))
-          nil
-          ;; FIXME: Should we respect *INSPECT-LENGTH* here? If not, what
-          ;; does *INSPECT-LENGTH* mean?
-          (coerce object 'list)))
+  (let ((length (min (length object) *inspect-length*)))
+    (values (format nil
+                    "The object is a ~:[~;displaced ~](VECTOR ~a) of length ~W.~%"
+                    (and (array-header-p object)
+                         (%array-displaced-p object))
+                    (array-element-type object)
+                    (length object))
+            nil
+            (unless (typep object '(simple-array nil))
+             (coerce (subseq object 0 length) 'list)))))
 
 (defun inspected-index-string (index rev-dimensions)
   (if (null rev-dimensions)
@@ -281,19 +270,18 @@ evaluated expressions.
                                       :displaced-to object))
          (dimensions (array-dimensions object))
          (reversed-elements nil))
-    ;; FIXME: Should we respect *INSPECT-LENGTH* here? If not, what does
-    ;; *INSPECT-LENGTH* mean?
-    (dotimes (i length)
-      (push (cons (format nil
-                          "~A "
-                          (inspected-index-string i (reverse dimensions)))
-                  (aref reference-array i))
-            reversed-elements))
-    (values (format nil "The object is ~:[a displaced~;an~] ARRAY of ~A.~%~
-                         Its dimensions are ~S.~%"
-                    (array-element-type object)
+    (unless (typep object '(simple-array nil))
+      (dotimes (i length)
+        (push (cons (format nil
+                            "~A "
+                            (inspected-index-string i (reverse dimensions)))
+                    (aref reference-array i))
+              reversed-elements)))
+    (values (format nil "The object is ~:[an~;a displaced~] ARRAY of ~A.~%~
+                         Its dimensions are ~:S.~%"
                     (and (array-header-p object)
                          (%array-displaced-p object))
+                    (array-element-type object)
                     dimensions)
             t
             (nreverse reversed-elements))))
