@@ -94,7 +94,7 @@
     ((head-var tail-var &optional user-head-var) &body body)
   (let ((l (and user-head-var (list (list user-head-var nil)))))
     `(let* ((,head-var (list nil)) (,tail-var ,head-var) ,@l)
-       (declare (truly-dynamic-extent ,head-var)
+       (declare (dynamic-extent ,head-var)
                 ,@(and user-head-var `((list ,user-head-var))))
        ,@body)))
 
@@ -121,8 +121,10 @@
                  (setq ncdrs (- (length (cdr form)) 2))))))
       (let ((answer
               (cond ((null ncdrs)
-                     `(when (setf (cdr ,tail-var) ,tail-form)
-                        (setq ,tail-var (last (cdr ,tail-var)))))
+                     (if (typep tail-form '(cons (eql copy-list)))
+                         `(setf ,tail-var (sb-impl::copy-list-to ,(second tail-form)  ,tail-var))
+                         `(when (setf (cdr ,tail-var) ,tail-form)
+                            (setq ,tail-var (last (cdr ,tail-var))))))
                     ((< ncdrs 0) (return-from loop-collect-rplacd nil))
                     ((= ncdrs 0)
                      ;; @@@@ Here we have a choice of two idioms:
@@ -290,10 +292,10 @@ code to be loaded.
              (:constructor !make-loop-universe)
              (:copier nil)
              (:predicate nil))
-  keywords             ; hash table, value = (fn-name . extra-data)
-  iteration-keywords   ; hash table, value = (fn-name . extra-data)
-  for-keywords         ; hash table, value = (fn-name . extra-data)
-  path-keywords)       ; hash table, value = (fn-name . extra-data)
+  (keywords nil :read-only t)             ; hash table, value = (fn-name . extra-data)
+  (iteration-keywords nil :read-only t)   ; hash table, value = (fn-name . extra-data)
+  (for-keywords nil :read-only t)         ; hash table, value = (fn-name . extra-data)
+  (path-keywords nil :read-only t))       ; hash table, value = (fn-name . extra-data)
 (declaim (sb-ext:freeze-type loop-universe))
 (defmethod print-object ((u loop-universe) stream)
   (print-unreadable-object (u stream :type t :identity t)))
@@ -562,8 +564,7 @@ code to be loaded.
   (do ((l (source-context loop) (cdr l)) (new nil (cons (car l) new)))
       ((eq l (cdr (source-code loop))) (nreverse new))))
 
-(defun loop-error (format-string &rest format-args)
-  #-sb-xc-host(declare (optimize allow-non-returning-tail-call))
+(define-error-wrapper loop-error (format-string &rest format-args)
   (%program-error "~?~%current LOOP context:~{ ~S~}."
                   format-string format-args (loop-context)))
 
@@ -1073,11 +1074,11 @@ code to be loaded.
 (defstruct (loop-collector
             (:copier nil)
             (:predicate nil))
-  name
-  class
+  (name nil :read-only t)
+  (class nil :read-only t)
   (history nil)
   (tempvars nil)
-  specified-type
+  (specified-type nil :read-only t)
   dtype
   (data nil)) ;collector-specific data
 (declaim (sb-ext:freeze-type loop-collector))
@@ -1368,7 +1369,7 @@ code to be loaded.
 
 (defun loop-for-across (var val data-type)
   (loop-make-var var nil data-type)
-  (let ((vector-var (sb-xc:gensym "LOOP-ACROSS-VECTOR-"))
+  (let ((vector-var (gensym "LOOP-ACROSS-VECTOR-"))
         (index-var (gensym "LOOP-ACROSS-INDEX-")))
     (multiple-value-bind (vector-form constantp vector-value)
         (loop-constant-fold-if-possible val 'vector)
@@ -1436,24 +1437,28 @@ code to be loaded.
                               `(,first-endtest ,step () ,pseudo)))))))))))
 
 (defun loop-for-in (var val data-type)
-  (multiple-value-bind (list constantp list-value)
-      (loop-constant-fold-if-possible val)
-    (let ((listvar (gensym "LOOP-LIST-")))
-      (loop-make-var var nil data-type)
-      (loop-make-var listvar
-                     ;; Don't want to assert the type, as ENDP will do that
-                     `(the* (list :use-annotations t :source-form ,list) ,list)
-                     t)
-      (let ((list-step (loop-list-step listvar)))
-        (let* ((first-endtest `(endp ,listvar))
-               (other-endtest first-endtest)
-               (step `(,var (car ,listvar)))
-               (pseudo-step `(,listvar ,list-step)))
-          (when (and constantp (listp list-value))
-            (setq first-endtest (null list-value)))
-          `(,other-endtest ,step () ,pseudo-step
-            ,@(and (neq first-endtest other-endtest)
-                   `(,first-endtest ,step () ,pseudo-step))))))))
+  (if (and (typep val '(cons (eql reverse) (cons t null)))
+           (not (sb-c::fun-lexically-notinline-p 'reverse
+                                                 (macro-environment *loop*))))
+      (loop-for-across var `(list-reverse-into-vector ,(second val)) data-type)
+      (multiple-value-bind (list constantp list-value)
+          (loop-constant-fold-if-possible val)
+        (let ((listvar (gensym "LOOP-LIST-")))
+          (loop-make-var var nil data-type)
+          (loop-make-var listvar
+                         ;; Don't want to assert the type, as ENDP will do that
+                         `(the* (list :use-annotations t :source-form ,list) ,list)
+                         t)
+          (let ((list-step (loop-list-step listvar)))
+            (let* ((first-endtest `(endp ,listvar))
+                   (other-endtest first-endtest)
+                   (step `(,var (car ,listvar)))
+                   (pseudo-step `(,listvar ,list-step)))
+              (when (and constantp (listp list-value))
+                (setq first-endtest (null list-value)))
+              `(,other-endtest ,step () ,pseudo-step
+                               ,@(and (neq first-endtest other-endtest)
+                                      `(,first-endtest ,step () ,pseudo-step)))))))))
 
 ;;;; iteration paths
 
@@ -1650,14 +1655,14 @@ code to be loaded.
             (setq endform (if limit-constantp
                               `',limit-value
                               (loop-make-var
-                                 (sb-xc:gensym "LOOP-LIMIT-") form
+                                 (gensym "LOOP-LIMIT-") form
                                  `(and ,indexv-type real)))))
            (:by
             (multiple-value-setq (form stepby-constantp stepby)
               (loop-constant-fold-if-possible form
                                               `(and ,indexv-type (real (0)))))
             (unless stepby-constantp
-              (loop-make-var (setq stepby (sb-xc:gensym "LOOP-STEP-BY-"))
+              (loop-make-var (setq stepby (gensym "LOOP-STEP-BY-"))
                  form
                  `(and ,indexv-type (real (0)))
                  t)))
