@@ -196,13 +196,18 @@ variable: an unreadable object representing the error is printed instead.")
            (%with-output-to-string (stream)
               (sb-pretty:output-pretty-object stream fun object))
            (let ((buffer-size (approx-chars-in-repr object)))
-             (let* ((string (make-string buffer-size :element-type 'base-char))
+             (let* ((string (make-string buffer-size :element-type 'base-char
+                                         :initial-element (code-char 0)))
                     (stream (%make-finite-base-string-output-stream string)))
                (declare (inline %make-finite-base-string-output-stream))
-               (declare (truly-dynamic-extent stream))
+               (declare (dynamic-extent stream))
                (output-integer object stream *print-base* *print-radix*)
-               (%shrink-vector string
-                               (finite-base-string-output-stream-pointer stream)))))))
+               ;; ASSUMPTION: we use pre-zeroed memory for unboxed objects.
+               ;; So we can avoid calling %SHRINK-VECTOR, and instead directly
+               ;; set the length.
+               (setf (%array-fill-pointer string)
+                     (finite-base-string-output-stream-pointer stream))
+               string)))))
     ;; Could do something for other numeric types, symbols, ...
     (t
      (%with-output-to-string (stream)
@@ -578,12 +583,11 @@ variable: an unreadable object representing the error is printed instead.")
         (return-from output-ugly-object
           (print-unreadable-object (object stream :identity t)
             (prin1 'instance stream))))
-      (let* ((wrapper (layout-friend layout))
-             (classoid (wrapper-classoid wrapper)))
+      (let ((classoid (layout-classoid layout)))
         ;; Additionally, don't crash if the object is an obsolete thing with
         ;; no update protocol.
         (when (or (sb-kernel::undefined-classoid-p classoid)
-                  (and (wrapper-invalid wrapper)
+                  (and (layout-invalid layout)
                        (logtest (layout-flags layout)
                                 (logior +structure-layout-flag+
                                         +condition-layout-flag+))))
@@ -1344,7 +1348,7 @@ variable: an unreadable object representing the error is printed instead.")
                       ;; the character.
                       (write-char (schar chars r) stream)))))
       (cond ((typep integer 'word) ; Division vops can handle this all inline.
-             #+(and gencgc c-stack-is-control-stack) ; strings can be DX
+             #+c-stack-is-control-stack ; strings can be DX-allocated
              ;; For bases exceeding 10 we know how many characters (at most)
              ;; will be output. This allows for a single %WRITE-STRING call.
              ;; There's diminishing payback for other bases because the fixed array
@@ -1355,10 +1359,10 @@ variable: an unreadable object representing the error is printed instead.")
                  (let* ((ptr #.(length (write-to-string sb-ext:most-positive-word
                                                         :base 10)))
                         (buffer (make-array ptr :element-type 'base-char)))
-                   (declare (truly-dynamic-extent buffer))
+                   (declare (dynamic-extent buffer))
                    (iterative-algorithm)
                    (%write-string buffer stream ptr (length buffer))))
-             #-(and gencgc c-stack-is-control-stack) ; strings can not be DX
+             #-c-stack-is-control-stack ; strings can't be DX-allocated
              ;; Use the alien stack, which is not as fast as using the control stack
              ;; (when we can). Even the absence of 0-fill doesn't make up for it.
              ;; Since we've no choice in the matter, might as well allow
@@ -1536,10 +1540,6 @@ variable: an unreadable object representing the error is printed instead.")
 ;;; paper might not bring immediate illumination as CSR has attempted
 ;;; to turn idiomatic Scheme into idiomatic Lisp.
 ;;;
-;;; FIXME: figure 1 from Burger and Dybvig is the unoptimized
-;;; algorithm, noticeably slow at finding the exponent.  Figure 2 has
-;;; an improved algorithm, but CSR ran out of energy.
-;;;
 ;;; possible extension for the enthusiastic: printing floats in bases
 ;;; other than base 10.
 (defconstant single-float-min-e
@@ -1575,28 +1575,37 @@ variable: an unreadable object representing the error is printed instead.")
         (let ((shift (- float-digits (integer-length f))))
           (setq f (ash f shift)
                 e (- e shift))))
-      (let ( ;; FIXME: these even tests assume normal IEEE rounding
+      (let (;; FIXME: these even tests assume normal IEEE rounding
             ;; mode.  I wonder if we should cater for non-normal?
             (high-ok (evenp f))
             (low-ok (evenp f)))
-        (labels ((scale (r s m+ m-)
-                   (do ((r+m+ (+ r m+))
-                        (k 0 (1+ k))
-                        (s s (* s print-base)))
-                       ((not (or (> r+m+ s)
-                                 (and high-ok (= r+m+ s))))
-                        (do ((k k (1- k))
-                             (r r (* r print-base))
-                             (m+ m+ (* m+ print-base))
-                             (m- m- (* m- print-base)))
-                            ((not (and (> r m-) ; Extension to handle zero
-                                       (let ((x (* (+ r m+) print-base)))
-                                         (or (< x s)
-                                             (and (not high-ok)
-                                                  (= x s))))))
-                             (funcall prologue-fun k)
-                             (generate r s m+ m-)
-                             (funcall epilogue-fun k))))))
+        (labels ((expt2 (n)
+                   (svref #.(coerce (loop for i from 0 to 972 collect (expt 2 i))
+                                    'vector) n))
+                 (expt10 (n)
+                   (svref #.(coerce (loop for i from 0 to 323 collect (expt 10 i))
+                                    'vector) n))
+                 (scale (r s m+ m-)
+                   (let ((est (truly-the (integer -323 309)
+                                         (ceiling (- (* (+ e (integer-length (truly-the sb-kernel:double-float-significand f)) -1)
+                                                        (log $2d0 10))
+                                                     $1.0e-10)))))
+                     (if (>= est 0)
+                         (fixup r (* s (expt10 est)) m+ m- est)
+                         (let ((scale (expt10 (- est))))
+                           (fixup (* r scale)
+                                  s (* m+ scale) (* m- scale) est)))))
+
+                 (fixup (r s m+ m- k)
+                   (let ((r+m+ (+ r m+)))
+                     (when (if high-ok
+                               (>= r+m+ s)
+                               (> r+m+ s))
+                       (incf k)
+                       (setf s (* s 10)))
+                     (funcall prologue-fun k)
+                     (generate r s m+ m-)
+                     (funcall epilogue-fun k)))
                  (generate (r s m+ m-)
                    (let (d tc1 tc2)
                      (tagbody
@@ -1621,65 +1630,81 @@ variable: an unreadable object representing the error is printed instead.")
                                    (t
                                     (1+ d)))))
                           (funcall char-fun d)))))
-                 (initialize ()
-                   (let (r s m+ m-)
-                     (cond ((>= e 0)
-                            (let ((be (expt float-radix e)))
-                              (if (/= f (expt float-radix (1- float-digits)))
-                                  ;; multiply F by 2 first, avoding consing two bignums
-                                  (setf r (* f 2 be)
-                                        s 2
-                                        m+ be
-                                        m- be)
-                                  (setf m- be
-                                        m+ (* be float-radix)
-                                        r (* f 2 m+)
-                                        s (* float-radix 2)))))
-                           ((or (= e min-e)
-                                (/= f (expt float-radix (1- float-digits))))
-                            (setf r (* f 2)
-                                  s (expt float-radix (- 1 e))
-                                  m+ 1
-                                  m- 1))
-                           (t
-                            (setf r (* f float-radix 2)
-                                  s (expt float-radix (- 2 e))
-                                  m+ float-radix
-                                  m- 1)))
-                     (when position
-                       (when relativep
-                         (aver (> position 0))
-                         (do ((k 0 (1+ k))
-                              ;; running out of letters here
-                              (l 1 (* l print-base)))
-                             ((>= (* s l) (+ r m+))
-                              ;; k is now \hat{k}
-                              (if (< (+ r (* s (/ (expt print-base (- k position)) 2)))
-                                     (* s l))
-                                  (setf position (- k position))
-                                  (setf position (- k position 1))))))
-                       (let* ((x (/ (* s (expt print-base position)) 2))
-                              (low (max m- x))
-                              (high (max m+ x)))
-                         (when (<= m- low)
-                           (setf m- low)
-                           (setf low-ok t))
-                         (when (<= m+ high)
-                           (setf m+ high)
-                           (setf high-ok t))))
-                     (values r s m+ m-))))
-          (multiple-value-bind (r s m+ m-) (initialize)
-            (scale r s m+ m-)))))))
+                 (scale-p (r s m+ m-)
+                   (when relativep
+                     (aver (> position 0))
+                     (do ((k 0 (1+ k))
+                          ;; running out of letters here
+                          (l 1 (* l print-base)))
+                         ((>= (* s l) (+ r m+))
+                          ;; k is now \hat{k}
+                          (if (< (+ r (* s (/ (expt print-base (- k position)) 2)))
+                                 (* s l))
+                              (setf position (- k position))
+                              (setf position (- k position 1))))))
+                   (let* ((x (/ (* s (expt print-base position)) 2))
+                          (low (max m- x))
+                          (high (max m+ x)))
+                     (when (<= m- low)
+                       (setf m- low)
+                       (setf low-ok t))
+                     (when (<= m+ high)
+                       (setf m+ high)
+                       (setf high-ok t)))
+                   (do ((r+m+ (+ r m+))
+                        (k 0 (1+ k))
+                        (s s (* s print-base)))
+                       ((not (or (> r+m+ s)
+                                 (and high-ok (= r+m+ s))))
+                        (do ((k k (1- k))
+                             (r r (* r print-base))
+                             (m+ m+ (* m+ print-base))
+                             (m- m- (* m- print-base)))
+                            ((not (and (> r m-) ; Extension to handle zero
+                                       (let ((x (* (+ r m+) print-base)))
+                                         (or (< x s)
+                                             (and (not high-ok)
+                                                  (= x s))))))
+                             (funcall prologue-fun k)
+                             (generate r s m+ m-)
+                             (funcall epilogue-fun k)))))))
+          (let (r s m+ m-)
+            (cond ((>= e 0)
+                   (let ((be (expt2 e)))
+                     (setf m- be)
+                     (if (/= f (expt float-radix (1- float-digits)))
+                         (setf r (ash f (+ e 1))
+                               s 2
+                               m+ be)
+                         (setf m+ (expt2 (1+ e))
+                               r (ash f (+ e 2))
+                               s (* float-radix 2)))))
+                  ((or (= e min-e)
+                       (/= f (expt float-radix (1- float-digits))))
+                   (setf r (* f 2)
+                         s (expt float-radix (- 1 e))
+                         m+ 1
+                         m- 1))
+                  (t
+                   (setf r (* f float-radix 2)
+                         s (expt float-radix (- 2 e))
+                         m+ float-radix
+                         m- 1)))
+            (if position
+                (scale-p r s m+ m-)
+                (scale r s m+ m-))))))))
 
 (defun flonum-to-digits (float &optional position relativep)
-  (let ((digit-characters "0123456789"))
-    (with-push-char (:element-type base-char)
-      (%flonum-to-digits
-       (lambda (d)
-         (push-char (char digit-characters d)))
-       (lambda (k) k)
-       (lambda (k) (values k (get-pushed-string)))
-       float position relativep))))
+  (if (zerop float)
+      (values 0 "0")
+      (let ((digit-characters "0123456789"))
+        (with-push-char (:element-type base-char)
+          (%flonum-to-digits
+           (lambda (d)
+             (push-char (char digit-characters d)))
+           (lambda (k) k)
+           (lambda (k) (values k (get-pushed-string)))
+           float position relativep)))))
 
 (defun print-float (float stream)
   (let ((position 0)
@@ -1844,20 +1869,14 @@ variable: an unreadable object representing the error is printed instead.")
            (write-string (if (float-trapping-nan-p x) " trapping" " quiet") stream)
            (write-string " NaN" stream))))
     (t
-     (let ((x (cond ((minusp (float-sign x))
-                     (write-char #\- stream)
-                     (- x))
-                    (t
-                     x))))
-       (cond
-         ((zerop x)
-          (write-string "0.0" stream)
-          (print-float-exponent x 0 stream))
-         (t
-          (print-float x stream)))))))
-
-
-
+     (when (plusp (float-sign-bit x))
+       (write-char #\- stream))
+     (cond
+       ((zerop x)
+        (write-string "0.0" stream)
+        (print-float-exponent x 0 stream))
+       (t
+        (print-float x stream))))))
 
 ;;;; other leaf objects
 
@@ -1869,9 +1888,14 @@ variable: an unreadable object representing the error is printed instead.")
                            (standard-char-p char)))
             (name (char-name char)))
         (write-string "#\\" stream)
-        (if (and name (or (not graphicp) *print-readably*))
-            (quote-string name stream)
-            (write-char char stream)))
+        (cond
+          ((and name (or (not graphicp) *print-readably*)) (quote-string name stream))
+          (t
+           (write-char char stream)
+           ;; KLUDGE: arguably this should be in an :AROUND method
+           ;; specialized on ((EQL #\SPACE) T).
+           (when (and (eql char #\Space))
+             (sb-pretty:note-significant-space stream)))))
       (write-char char stream)))
 
 (defmethod print-object ((sap system-area-pointer) stream)
@@ -1881,6 +1905,20 @@ variable: an unreadable object representing the error is printed instead.")
          (print-unreadable-object (sap stream)
            (format stream "system area pointer: #X~8,'0X" (sap-int sap))))))
 
+#+weak-vector-readbarrier
+(defmethod print-object ((self weak-pointer) stream)
+  (let ((vectorp (weak-vector-p self)))
+    (print-unreadable-object (self stream :identity vectorp)
+      (if vectorp
+          (format stream "weak array [~d]" (weak-vector-len self))
+          (multiple-value-bind (value validp) (weak-pointer-value self)
+            (cond (validp
+                   (write-string "weak pointer: " stream)
+                   (write value :stream stream))
+                  (t
+                   (write-string "broken weak pointer" stream))))))))
+
+#-weak-vector-readbarrier
 (defmethod print-object ((weak-pointer weak-pointer) stream)
   (print-unreadable-object (weak-pointer stream)
     (multiple-value-bind (value validp) (weak-pointer-value weak-pointer)
@@ -2067,7 +2105,7 @@ variable: an unreadable object representing the error is printed instead.")
          (let ((widetag (widetag-of object)))
            (case widetag
              (#.sb-vm:value-cell-widetag
-              (write-string "value cell " stream)
+              (write-string "value-cell " stream)
               (output-object (value-cell-ref object) stream))
              #+nil
              (#.sb-vm:filler-widetag
