@@ -77,7 +77,7 @@
     (def-accessor sap-ref-lispobj)
     (def-accessor sap-ref-single)
     (def-accessor sap-ref-double))
-  (def %byte-blt (src src-start dst dst-start dst-end))
+  (def %byte-blt (src src-start dst dst-start count))
   (def shift-towards-start (number count))
   (def shift-towards-end (number count))
   (def get-header-data)
@@ -102,6 +102,7 @@
   #-untagged-fdefns (def code-header-ref (code-obj index))
   (def %vector-raw-bits (object offset))
   (def %set-vector-raw-bits (object offset value))
+  #-weak-vector-readbarrier (def weak-vector-len)
   (def single-float-bits)
   (def double-float-high-bits)
   #+64-bit
@@ -110,6 +111,7 @@
   (def value-cell-ref)
   (def %caller-frame ())
   (def %caller-pc ())
+  (def %code-debug-info)
   (def sb-vm::%code-fixups)
   #+x86-64 (def pointerp)
   (def sb-bignum:%bignum-length)
@@ -119,20 +121,19 @@
   (def %make-instance/mixed)
   (def %instance-length) ; Given an instance, return its length.
   (def %instance-layout)
-  (def %instance-wrapper)
   (def %set-instance-layout (instance new-value))
   ; (def %instance-ref (instance index)) ; defined in 'target-defstruct'
   (def %instance-set (instance index new-value))
   ;; funcallable instances
-  (def %make-funcallable-instance)
+  ;(def %make-funcallable-instance)
   (def %fun-layout)
-  (def %fun-wrapper)
   (def %set-fun-layout (fin new-value))
   (def %funcallable-instance-fun)
   (def (setf %funcallable-instance-fun) (fin new-value))
   (def %funcallable-instance-info (fin i))
-  #+compact-instance-header (progn (def wrapper-of)
+  #+compact-instance-header (progn (def layout-of)
                                    (def %instanceoid-layout))
+  #+64-bit (def layout-depthoid)
 
   ;; lists
   (def %rplaca (x val))
@@ -141,6 +142,11 @@
   #+compare-and-swap-vops
   (def* (%array-atomic-incf/word (array index diff))
         (%raw-instance-atomic-incf/word (instance index diff)))
+  #+(or x86 x86-64)
+  (def* (sb-vm::%cpu-identification (arg1 arg2))
+        (sb-vm::%vector-cas-pair (vector index old1 old2 new1 new2))
+        (sb-vm::%instance-cas-pair (instance index old1 old2 new1 new2))
+        (sb-vm::%cons-cas-pair (cons old1 old2 new1 new2)))
 
   #+sb-simd-pack
   (def* (%make-simd-pack (tag low high))
@@ -159,13 +165,9 @@
   (def current-fp ())
   (def stack-ref (s n))
   (def fun-code-header)
+  (def symbol-package-id)
   (def symbol-hash)
-  (def sb-vm::symbol-extra)
-  #+sb-thread (def symbol-tls-index)
   (def symbol-%info) ; primitive reader always needs a stub
-  ;; but the "wrapped" reader might not need a stub.
-  ;; If it's already a proper function, then it doesn't.
-  #.(if (fboundp 'symbol-dbinfo) (values) '(def symbol-dbinfo))
   #-(or x86 x86-64) (def lra-code-header)
   (def %make-lisp-obj)
   (def get-lisp-obj-address)
@@ -196,17 +198,6 @@
   "Hints the processor that the current thread is spin-looping."
   (spin-loop-hint))
 
-;;; The stub for sb-c::%structure-is-a should really use layout-id in the same way
-;;; that the vop does, however, because the all 64-bit architectures other than
-;;; x86-64 need to use with-pinned-objects to extract a layout-id, it is cheaper not to.
-;;; I should add a vop for uint32 access to raw slots.
-(defun sb-c::%structure-is-a (object-layout test-layout)
-  (or (eq object-layout test-layout)
-      (let ((depthoid (wrapper-depthoid test-layout))
-            (inherits (wrapper-inherits object-layout)))
-        (and (> (length inherits) depthoid)
-             (eq (svref inherits depthoid) test-layout)))))
-
 (defun %other-pointer-subtype-p (x choices)
   (and (%other-pointer-p x)
        (member (%other-pointer-widetag x) choices)
