@@ -17,9 +17,6 @@
 (declaim (maybe-inline upper-case-p lower-case-p both-case-p
                        digit-char-p))
 
-(deftype char-code ()
-  `(integer 0 (,char-code-limit)))
-
 (declaim (inline pack-3-codepoints))
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (defun pack-3-codepoints (first &optional (second 0) (third 0))
@@ -60,7 +57,7 @@
 
 (macrolet ((frob ()
              (flet ((file (name type)
-                      (sb-cold:find-bootstrap-file (format nil "output/~A.~A" name type)))
+                      (sb-cold:find-bootstrap-file (format nil "output/ucd/~A.~A" name type)))
                     (read-ub8-vector (pathname)
                       (with-open-file (stream pathname
                                               :element-type '(unsigned-byte 8))
@@ -88,10 +85,8 @@
                      (ucd-high-pages (read-ub8-vector (file "ucdhigh" "dat")))
                      (ucd-low-pages (read-ub8-vector (file "ucdlow" "dat")))
                      (decompositions (read-ub8-vector (file "decomp" "dat")))
-                     (primary-compositions (read-ub8-vector (file "comp" "dat")))
                      (case-data (read-ub8-vector (file "case" "dat")))
                      (case-pages (read-ub8-vector (file "casepages" "dat")))
-                     (collations (read-ub8-vector (file "collation" "dat")))
                      (high-pages (make-ubn-vector ucd-high-pages 2))
                      (low-pages (make-ubn-vector ucd-low-pages 2))
                      (%*character-case-pages*% (make-ubn-vector case-pages 1)))
@@ -103,19 +98,6 @@
                     (defconstant-eqx sb-unicode::+character-decompositions+
                         ,(make-ubn-vector decompositions 3) #'equalp)
                     (defconstant-eqx +character-case-pages+ ,%*character-case-pages*% #'equalp)
-                    (declaim (hash-table **character-primary-compositions**))
-                    (define-load-time-global **character-primary-compositions**
-                        ,(let ((info (make-ubn-vector primary-compositions 3))
-                               (alist))
-                           (dotimes (i (/ (length info) 3))
-                             (let* ((3i (* 3 i))
-                                    (key (dpb (aref info 3i) (byte 21 21) (aref info (1+ 3i))))
-                                    (value (aref info (+ 3i 2))))
-                               (push (cons key value) alist)))
-                           `(%stuff-hash-table (make-hash-table #+64-bit :test #+64-bit #'eq
-                                                                :size ,(/ (length info) 3))
-                                               ,(nreverse (coerce alist 'vector)))))
-
                     ,@(let* ((unicode-table
                                  (make-array
                                   (* 64 (1+ (aref %*character-case-pages*%
@@ -143,72 +125,37 @@
                                              (t
                                               (dotimes (i len)
                                                 (push (read-codepoint) ret))
-                                              (nreverse ret))))))
-                            (loop until (>= index length)
-                                  for key = (read-codepoint)
-                                  for upper = (read-length-tagged)
-                                  for lower = (read-length-tagged)
-                                  for page = (aref %*character-case-pages*% (ash key -6))
-                                  for i = (+ (ash page 6) (ldb (byte 6 0) key))
-                                  do
-                                  (setf (aref unicode-table i)
-                                        (if (or (consp upper)
-                                                (consp lower))
-                                            (cons upper lower)
-                                            (dpb upper (byte 21 21) lower)))
-                                  when
-                                  (flet (#+sb-unicode
-                                         (both-case-p-local (code)
-                                           (logbitp 7 (aref misc-database
-                                                            (+ 5 (misc-index-from-char-code
-                                                                  code high-pages low-pages))))))
-                                    (and (atom upper)
-                                         (atom lower)
-                                         ;; Some characters are only equal under unicode rules,
-                                         ;; e.g. #\MICRO_SIGN and #\GREEK_CAPITAL_LETTER_MU
-                                         #+sb-unicode
-                                         (both-case-p-local lower)
-                                         #+sb-unicode
-                                         (both-case-p-local upper)))
-                                  do
-                                  (setf (aref table (* i 2)) lower
-                                        (aref table (1+ (* i 2))) upper)))
+                                              (nreverse ret)))))
+                                   #+sb-unicode
+                                   (both-case-p-local (code)
+                                     (logbitp 7 (aref misc-database
+                                                      (+ 5 (misc-index-from-char-code
+                                                            code high-pages low-pages))))))
+                            (loop
+                              until (>= index length)
+                              do (let* ((key (read-codepoint))
+                                        (upper (read-length-tagged))
+                                        (lower (read-length-tagged))
+                                        (page (aref %*character-case-pages*% (ash key -6)))
+                                        (i (+ (ash page 6) (ldb (byte 6 0) key))))
+                                   (setf (aref unicode-table i)
+                                         (if (or (consp upper)
+                                                 (consp lower))
+                                             (cons upper lower)
+                                             (dpb upper (byte 21 21) lower)))
+                                   (when (and (atom upper)
+                                              (atom lower)
+                                              ;; Some characters are only equal under unicode rules,
+                                              ;; e.g. #\MICRO_SIGN and #\GREEK_CAPITAL_LETTER_MU
+                                              #+sb-unicode
+                                              (both-case-p-local lower)
+                                              #+sb-unicode
+                                              (both-case-p-local upper))
+                                     (setf (aref table (* i 2)) lower
+                                           (aref table (1+ (* i 2))) upper)))))
                           `((defconstant-eqx +character-unicode-cases+ ,unicode-table #'equalp)
                             (defconstant-eqx +character-cases+ ,table #'equalp)))
 
-                    (define-load-time-global **character-collations**
-                             ,(let* ((n-entries
-                                         (with-open-file (s (file "n-collation-entries" "lisp-expr"))
-                                           (read s)))
-                                     (table)
-                                     (index 0)
-                                     (info (make-ubn-vector collations 4))
-                                     (len (length info)))
-                                (loop while (< index len) do
-                                      (let* ((entry-head (aref info index))
-                                             (cp-length (ldb (byte 4 28) entry-head))
-                                             (key-length (ldb (byte 5 23) entry-head))
-                                             (key 0)
-                                             (codepoints nil))
-                                        (aver (and (/= cp-length 0) (/= key-length 0)))
-                                        (loop repeat cp-length do
-                                              (push (dpb 0 (byte 10 22) (aref info index))
-                                                    codepoints)
-                                              (incf index))
-                                        (setf codepoints (nreverse codepoints))
-                                        (dotimes (i key-length)
-                                          (setf (ldb (byte 32 (* i 32)) key) (aref info index))
-                                          (incf index))
-                                        ;; verify the validity of
-                                        ;; :test 'eq on 64-bit
-                                        #+64-bit (aver (sb-xc:typep (apply #'pack-3-codepoints codepoints)
-                                                                    'sb-xc:fixnum))
-                                        (push (cons (apply #'pack-3-codepoints codepoints) key) table)))
-                                (aver (= (length table) n-entries))
-                                `(%stuff-hash-table
-                                  (make-hash-table :size ,n-entries #+64-bit :test #+64-bit #'eq)
-                                  ,(nreverse (coerce table 'vector)))))
-
                     ,@(with-open-file
                          (stream (file "ucd-names" "lisp-expr"))
                        (with-open-file (u1-stream (file "ucd1-names" "lisp-expr"))
@@ -361,7 +308,7 @@
 
 ;;; The character database is made of several arrays.
 ;;; +CHARACTER-MISC-DATABASE+ is an array of bytes that encode character
-;;; attributes. Each entry in the misc database is +misc-width+ (currently 9)
+;;; attributes. Each entry in the misc database is +MISC-WIDTH+ (currently 9)
 ;;; bytes wide. Within each entry, the bytes represent: general category, BIDI
 ;;; class, canonical combining class, digit value, decomposition info, other
 ;;; flags, script, line break class, and age, respectively. Several of the
@@ -374,28 +321,30 @@
 ;;; and also sets its high bit if the decompositions are compatibility
 ;;; decompositions. The other flags byte encodes boolean properties. Bit 7 is
 ;;; set if the entry's characters are BOTH-CASE-P in the Common Lisp sense. Bit
-;;; 6 is set if the entry's characters hav a defined case transformation in
+;;; 6 is set if the entry's characters have a defined case transformation in
 ;;; Unicode. Bit 5 is set if the characters have the property BIDI_Mirrored=Y.
 ;;; Bits 3-0 encode the entry's East Asian Width. Bit 4 is unused. Age stores
 ;;; the minor version in bits 0-2, and the major version in the remaining 5
 ;;; bits.
 ;;;
-;;; To find which entry in +CHARACTER-MISC-DATABASE+ encodes a character's
-;;; attributes, first index +CHARACTER-HIGH-PAGES+ (an array of 16-bit
-;;; values) with the high 13 bits of the character's codepoint. If the result
-;;; value has its high bit set, the character is in a "compressed page". To
-;;; find the misc entry number, simply clear the high bit. If the high bit is
-;;; not set, the misc entry number must be looked up in
-;;; +CHARACTER-LOW-PAGES+, which is an array of 16-bit values. Each entry in
-;;; the array consists of two such values, the misc entry number and the
-;;; decomposition index. To find the misc entry number, index into
-;;; +CHARACTER-LOW-PAGES+ using the value retreived from
-;;; +CHARACTER-HIGH-PAGES+ (shifted left 8 bits) plus the low 8 bits of the
-;;; codepoint, all times two to account for the widtth of the entries. The
-;;; value in +CHARACTER-LOW-PAGES+ at this point is the misc entry number. To
-;;; transform a misc entry number into an index into
-;;; +CHARACTER-MISC-DATABASE+, multiply it by +misc-width*. This gives the
-;;; index of the start of the charater's misc entry in
+;;; To find which entry in +CHARACTER-MISC-DATABASE+ encodes a
+;;; character's attributes, first index +CHARACTER-HIGH-PAGES+ (an
+;;; array of 16-bit values) with the high 13 bits of the character's
+;;; codepoint. If the result value has its high bit set, the character
+;;; is in a "compressed page", where all characters on that
+;;; 256-character page have the same misc entry. To find the misc
+;;; entry number, simply clear the high bit. If the high bit is not
+;;; set, the misc entry number must be looked up in
+;;; +CHARACTER-LOW-PAGES+, which is an array of 16-bit values. Each
+;;; entry in the array consists of two such values, the misc entry
+;;; number and the decomposition index. To find the misc entry number,
+;;; index into +CHARACTER-LOW-PAGES+ using the value retreived from
+;;; +CHARACTER-HIGH-PAGES+ (shifted left 8 bits) plus the low 8 bits
+;;; of the codepoint, all times two to account for the widtth of the
+;;; entries. The value in +CHARACTER-LOW-PAGES+ at this point is the
+;;; misc entry number. To transform a misc entry number into an index
+;;; into +CHARACTER-MISC-DATABASE+, multiply it by +MISC-WIDTH+. This
+;;; gives the index of the start of the charater's misc entry in
 ;;; +CHARACTER-MISC-DATABASE+.
 ;;;
 ;;; To look up a character's decomposition, first retreive its
@@ -422,9 +371,9 @@
 ;;; a flat array twice as large, and it includes only the standard casing rules,
 ;;; so there's always just two characters.
 ;;;
-;;; Similarly, composition information is stored in **CHARACTER-COMPOSITIONS**,
-;;; which is a hash table of codepoints indexed by (+ (ash codepoint1 21)
-;;; codepoint2).
+;;; Primary composition information is stored in a hash table local to
+;;; PRIMARY-COMPOSITION, with (+ (ash codepoint1 21) codepoint2) as
+;;; keys and the composition as the value
 
 (declaim (ftype (sfunction (t) (unsigned-byte 16)) misc-index))
 (defun misc-index (char)
@@ -460,7 +409,6 @@ there are no character bits or fonts.)"
   "Coerce OBJECT into a CHARACTER if possible. Legal inputs are characters,
 strings and symbols of length 1."
   (flet ((do-error (control args)
-           (declare (optimize allow-non-returning-tail-call))
            (error 'simple-type-error
                   :datum object
                   ;;?? how to express "symbol with name of length 1"?
@@ -552,7 +500,7 @@ argument is an alphabetic character, A-Z or a-z; otherwise NIL."
 
 (defmacro with-case-info ((char index-var cases-var
                            &key miss-value
-                                (cases '+character-cases+))
+                                (cases +character-cases+))
                           &body body)
   (let ((code-var (gensym "CODE"))
         (shifted-var (gensym "SHIFTED"))
@@ -562,9 +510,9 @@ argument is an alphabetic character, A-Z or a-z; otherwise NIL."
            (declare (optimize (sb-c:insert-array-bounds-checks 0)))
          (let ((,code-var (char-code ,char)))
            (let* ((,shifted-var (ash ,code-var -6))
-                  (,page-var (if (>= ,shifted-var (length +character-case-pages+))
+                  (,page-var (if (>= ,shifted-var ,(length +character-case-pages+))
                                  (return ,miss-value)
-                                 (aref +character-case-pages+ ,shifted-var))))
+                                 (aref ,+character-case-pages+ ,shifted-var))))
              (if (= ,page-var 255)
                  ,miss-value
                  (let ((,index-var (* (+ (ash ,page-var 6)
@@ -602,7 +550,7 @@ lowercase eszet (U+DF)."
     (let ((code (aref cases (1+ index))))
       (if (zerop code)
           char
-          (code-char (truly-the char-code code))))))
+          (code-char (truly-the %char-code code))))))
 
 (defun char-downcase (char)
   "Return CHAR converted to lower-case if that is possible."
@@ -611,7 +559,7 @@ lowercase eszet (U+DF)."
     (let ((code (aref cases index)))
       (if (zerop code)
           char
-          (code-char (truly-the char-code code))))))
+          (code-char (truly-the %char-code code))))))
 
 (defun alphanumericp (char)
   "Given a character-object argument, ALPHANUMERICP returns T if the argument
@@ -643,11 +591,11 @@ is either numeric or alphabetic."
          (shifted (ash code -6))
          (page (if (>= shifted (length +character-case-pages+))
                    (return-from equal-char-code code)
-                   (aref +character-case-pages+ shifted))))
+                   (aref #.+character-case-pages+ shifted))))
     (if (= page 255)
         code
         (let ((down-code
-                (aref +character-cases+
+                (aref #.+character-cases+
                       (* (+ (ash page 6)
                             (ldb (byte 6 0) code))
                          2))))
@@ -765,21 +713,23 @@ Case is ignored." t))
 (defun digit-char-p (char &optional (radix 10.))
   "If char is a digit in the specified radix, returns the fixnum for which
 that digit stands, else returns NIL."
-  (if (<= (char-code char) 127)
-      (let ((weight (- (char-code char) 48)))
-        (cond ((minusp weight) nil)
-              ((<= radix 10.)
-               ;; Special-case ASCII digits in decimal and smaller radices.
-               (if (< weight radix) weight nil))
-              ;; Digits 0 - 9 are used as is, since radix is larger.
-              ((< weight 10) weight)
-              ;; Check for upper case A - Z.
-              ((and (>= (decf weight 7) 10) (< weight radix)) weight)
-              ;; Also check lower case a - z.
-              ((and (>= (decf weight 32) 10) (< weight radix)) weight)))
-      (let ((number (ucd-decimal-digit char)))
-        (when (and number (< (truly-the fixnum number) radix))
-          number))))
+  (let ((code (char-code char)))
+    (if (<= code 1632) ;; (loop for code from 127 when (digit-char-p (code-char code)) return code)
+        (let ((weight (- code 48)))
+          (cond ((minusp weight) nil)
+                ((<= radix 10.)
+                 ;; Special-case ASCII digits in decimal and smaller radices.
+                 (if (< weight radix) weight nil))
+                ;; Digits 0 - 9 are used as is, since radix is larger.
+                ((< weight 10) weight)
+                (t
+                 (let ((weight (logior #x20 code))) ;; downcase ASCII characters.
+                   (when (and (>= (decf weight (- (char-code #\a) 10)) 10)
+                              (< weight radix))
+                     weight) ))))
+        (let ((number (ucd-decimal-digit char)))
+          (when (and number (< number radix))
+            number)))))
 
 (defun digit-char (weight &optional (radix 10))
   "All arguments must be integers. Returns a character object that represents
