@@ -182,7 +182,9 @@ corresponds to NAME, or NIL if there is none."
     (locally
         (declare (optimize (sb-c::float-accuracy 0)))
       (let ((result (alien-funcall (extern-alien "open" (function int c-string int &optional int))
-                                   path (logior flags) mode)))
+                                   path (logior flags
+                                                #+largefile o_largefile)
+                                   mode)))
         (if (minusp result)
             (values nil (get-errno))
             (values result 0))))))
@@ -549,7 +551,7 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 (defun unix-ioctl (fd cmd arg)
   (declare (type unix-fd fd)
            (type word cmd))
-  (void-syscall ("ioctl" int unsigned-long (* char)) fd cmd arg))
+  (void-syscall ("ioctl" int unsigned-long &optional (* char)) fd cmd arg))
 
 ;;;; sys/resource.h
 
@@ -656,8 +658,6 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
      (declare (type (integer 0 #.fd-setsize) ,n))
      ,@body))
 
-;;;; FIXME: Why have both UNIX-SELECT and UNIX-FAST-SELECT?
-
 ;;; Perform the UNIX select(2) system call.
 (declaim (inline unix-fast-select))
 (defun unix-fast-select (num-descriptors
@@ -683,69 +683,7 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
                (note-dangerous-wait "select(2)"))
              (select (int-sap 0)))))))
 
-;;; UNIX-SELECT accepts sets of file descriptors and waits for an event
-;;; to happen on one of them or to time out.
-(declaim (inline num-to-fd-set fd-set-to-num))
-(defun num-to-fd-set (fdset num)
-  (typecase num
-    (fixnum
-     (setf (deref (slot fdset 'fds-bits) 0) num)
-     (loop for index from 1 below (/ fd-setsize
-                                     sb-vm:n-machine-word-bits)
-           do (setf (deref (slot fdset 'fds-bits) index) 0)))
-    (t
-     (loop for index from 0 below (/ fd-setsize
-                                     sb-vm:n-machine-word-bits)
-           do (setf (deref (slot fdset 'fds-bits) index)
-                    (ldb (byte sb-vm:n-machine-word-bits
-                               (* index sb-vm:n-machine-word-bits))
-                         num))))))
-
-(defun fd-set-to-num (nfds fdset)
-  (if (<= nfds sb-vm:n-machine-word-bits)
-      (deref (slot fdset 'fds-bits) 0)
-      (loop for index below (/ fd-setsize
-                               sb-vm:n-machine-word-bits)
-            sum (ash (deref (slot fdset 'fds-bits) index)
-                     (* index sb-vm:n-machine-word-bits)))))
-
-;;; Examine the sets of descriptors passed as arguments to see whether
-;;; they are ready for reading and writing. See the UNIX Programmer's
-;;; Manual for more information.
-(defun unix-select (nfds rdfds wrfds xpfds to-secs &optional (to-usecs 0))
-  (declare (muffle-conditions compiler-note))
-  (declare (type integer nfds)
-           (type unsigned-byte rdfds wrfds xpfds)
-           (type (or (unsigned-byte 31) null) to-secs)
-           (type (unsigned-byte 31) to-usecs)
-           (optimize (speed 3) (safety 0)))
-  (with-fd-setsize (nfds)
-    (with-alien ((tv (struct timeval))
-                 (rdf (struct fd-set))
-                 (wrf (struct fd-set))
-                 (xpf (struct fd-set)))
-      (cond (to-secs
-             (setf (slot tv 'tv-sec) to-secs
-                   (slot tv 'tv-usec) to-usecs))
-            ((not *interrupts-enabled*)
-             (note-dangerous-wait "select(2)")))
-      (num-to-fd-set rdf rdfds)
-      (num-to-fd-set wrf wrfds)
-      (num-to-fd-set xpf xpfds)
-      (macrolet ((frob (lispvar alienvar)
-                   `(if (zerop ,lispvar)
-                        (int-sap 0)
-                        (alien-sap (addr ,alienvar)))))
-        (syscall ("sb_select" int (* (struct fd-set)) (* (struct fd-set))
-                           (* (struct fd-set)) (* (struct timeval)))
-                 (values result
-                         (fd-set-to-num nfds rdf)
-                         (fd-set-to-num nfds wrf)
-                         (fd-set-to-num nfds xpf))
-                 nfds (frob rdfds rdf) (frob wrfds wrf) (frob xpfds xpf)
-                 (if to-secs (alien-sap (addr tv)) (int-sap 0)))))))
-
-;;; Lisp-side implmentations of FD_FOO macros.
+;;; Lisp-side implementations of FD_FOO macros.
 (declaim (inline fd-set fd-clr fd-isset fd-zero))
 (defun fd-set (offset fd-set)
   (multiple-value-bind (word bit) (floor offset
@@ -932,8 +870,9 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 
 (define-alien-routine get-timezone int
   (when time-t)
-  ;; KLUDGE: the runtime `boolean' is defined as `int', but the alien
-  ;; type is N-WORD-BITS wide.
+  ;; BOOLEAN is N-WORD-BITS normally. Reduce it to an unsigned int in size.
+  ;; But we can't just put UNSIGNED-INT here because the clients of this function
+  ;; want to receive a T or NIL, not a 1 or 0.
   (daylight-savings-p (boolean 32) :out))
 #-win32
 (defun nanosleep (secs nsecs)
@@ -1048,7 +987,9 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 #-win32
 (progn
 
+  #-avoid-clock-gettime
   (declaim (inline clock-gettime))
+  #-avoid-clock-gettime
   (defun clock-gettime (clockid)
     (declare (type (signed-byte 32) clockid))
     (with-alien ((ts (struct timespec)))
@@ -1083,7 +1024,10 @@ the UNIX epoch (January 1st 1970.)"
           ;; By scaling down we end up with far less resolution than clock-realtime
           ;; offers, and COARSE is about twice as fast, so use that, but only for linux.
           ;; BSD has something similar.
+          #-avoid-clock-gettime
           (clock-gettime #+linux clock-monotonic-coarse #-linux clock-monotonic)
+          #+avoid-clock-gettime
+          (multiple-value-bind (c-sec c-usec) (get-time-of-day) (values c-sec (* c-usec 1000)))
 
         #+64-bit ;; I know that my math is valid for 64-bit.
         (declare (optimize (sb-c::type-check 0)))
@@ -1132,12 +1076,29 @@ the UNIX epoch (January 1st 1970.)"
                   current)))))))
 
   (declaim (inline system-internal-run-time))
-  #-sunos ; defined in sunos-os
+
+  ;; SunOS defines CLOCK_PROCESS_CPUTIME_ID but you get EINVAL if you try to use it,
+  ;; also use the same trick when clock_gettime should be avoided.
+  #-(or sunos avoid-clock-gettime)
   (defun system-internal-run-time ()
     (multiple-value-bind (sec nsec) (clock-gettime clock-process-cputime-id)
       (+ (* sec internal-time-units-per-second)
          (floor (+ nsec (floor nanoseconds-per-internal-time-unit 2))
-                nanoseconds-per-internal-time-unit)))))
+                nanoseconds-per-internal-time-unit))))
+  #+(or sunos avoid-clock-gettime)
+  (defun system-internal-run-time ()
+    (multiple-value-bind (utime-sec utime-usec stime-sec stime-usec)
+        (with-alien ((usage (struct sb-unix::rusage)))
+          (syscall* ("sb_getrusage" int (* (struct sb-unix::rusage)))
+                    (values (slot (slot usage 'sb-unix::ru-utime) 'sb-unix::tv-sec)
+                            (slot (slot usage 'sb-unix::ru-utime) 'sb-unix::tv-usec)
+                            (slot (slot usage 'sb-unix::ru-stime) 'sb-unix::tv-sec)
+                            (slot (slot usage 'sb-unix::ru-stime) 'sb-unix::tv-usec))
+                    rusage_self (addr usage)))
+      (+ (* (+ utime-sec stime-sec) internal-time-units-per-second)
+         (floor (+ utime-usec stime-usec
+                   (floor microseconds-per-internal-time-unit 2))
+                microseconds-per-internal-time-unit)))))
 
 ;;; FIXME, KLUDGE: GET-TIME-OF-DAY used to be UNIX-GETTIMEOFDAY, and had a
 ;;; primary return value indicating sucess, and also returned timezone
