@@ -58,11 +58,13 @@
 
 ;;; Signal a warning if appropriate and set *FOO-DETECTED*.
 (defun note-lossage (format-string &rest format-args)
+  (declare (explicit-check))
   (setq *lossage-detected* t)
   (when *lossage-fun*
     (apply *lossage-fun* format-string format-args))
   (values))
 (defun note-unwinnage (format-string &rest format-args)
+  (declare (explicit-check))
   (setq *unwinnage-detected* t)
   (when *unwinnage-fun*
     (apply *unwinnage-fun* format-string format-args))
@@ -237,6 +239,9 @@ and no value was provided for it." name))))))))))
 (defun check-arg-type (lvar type n)
   (declare (type lvar lvar) (type ctype type) (type index n))
   (cond
+    ((eq (lvar-type lvar) *empty-type*)
+     (note-unwinnage "The ~:R argument never returns a value." n)
+     nil)
     ((not *ctype-test-fun*))
     ((not (constant-type-p type))
      (let* ((ctype (lvar-type lvar))
@@ -246,9 +251,6 @@ and no value was provided for it." name))))))))))
                 (note-lossage "The ~:R argument is a ~S, not a ~S."
                               n (type-specifier ctype) (type-specifier type)))
               nil)
-             ((eq ctype *empty-type*)
-              (note-unwinnage "The ~:R argument never returns a value." n)
-              nil)
              (t t))))
     ((not (constant-lvar-p lvar))
      (note-unwinnage "The ~:R argument is not a constant." n)
@@ -434,6 +436,9 @@ and no value was provided for it." name))))))))))
 ;;; CALL. If TYPE is supplied and not null, then we merge the
 ;;; information into the information already accumulated in TYPE.
 (defun note-fun-use (call &optional type)
+  (declare (inline make-approximate-key-info
+                   make-approximate-fun-type)
+           (sb-c::tlab :system))
   (declare (type combination call)
            (type (or approximate-fun-type null) type)
            #-sb-xc-host
@@ -879,26 +884,6 @@ and no value was provided for it." name))))))))))
     ((t) nil)
     (t `(:not . ,explicit-check))))
 
-;;; FIXME: This is quite similar to ASSERT-NEW-DEFINITION.
-(defun assert-global-function-definition-type (name fun)
-  (declare (type functional fun))
-  (let ((where (info :function :where-from name))
-        (explicit-check (getf (functional-plist fun) 'explicit-check)))
-    (if (eq where :declared)
-        (let ((type
-               (massage-global-definition-type (global-ftype name) fun)))
-          (setf (leaf-type fun) type)
-          (assert-definition-type
-           fun type
-           :unwinnage-fun #'compiler-notify
-           :where "proclamation"
-           :really-assert (explicit-check->really-assert explicit-check)))
-        ;; Can't actually test this. DEFSTRUCTs declare this, but non-toplevel
-        ;; ones won't have an FTYPE at compile-time.
-        #+nil
-        (when explicit-check
-          (warn "Explicit-check without known FTYPE is meaningless")))))
-
 ;;; If the function has both &REST and &KEY, FIND-OPTIONAL-DISPATCH-TYPES
 ;;; doesn't complain about the type missing &REST -- which is good, because in
 ;;; that case &REST is really an implementation detail and not part of the
@@ -923,9 +908,10 @@ and no value was provided for it." name))))))))))
 ;;; Call FUN with (arg-lvar arg-type lvars &optional annotation)
 (defun map-combination-args-and-types (fun call &optional info
                                                           unknown-keys-fun
-                                                          declared-only)
+                                                          defined-here
+                                                          asserted-type)
   (declare (type function fun) (type combination call))
-  (binding* ((type (lvar-fun-type (combination-fun call) declared-only declared-only))
+  (binding* ((type (lvar-fun-type (combination-fun call) defined-here asserted-type))
              (nil (fun-type-p type) :exit-if-null)
              (annotation (and info
                               (fun-info-annotation info)))
@@ -992,7 +978,7 @@ and no value was provided for it." name))))))))))
         (use-lvar cast internal-lvar)
         t))))
 
-(defun apply-type-annotation (fun-name arg type lvars policy &optional annotation)
+(defun apply-type-annotation (fun-name arg type lvars policy &optional annotation context)
   (case (car annotation)
     (function-designator
      (assert-function-designator fun-name lvars arg (cdr annotation) policy)
@@ -1016,12 +1002,12 @@ and no value was provided for it." name))))))))))
         (add-annotation arg
                         (make-lvar-proper-sequence-annotation
                          :kind (car annotation)))))
-     (assert-lvar-type arg type policy))))
+     (assert-lvar-type arg type policy context))))
 
 ;;; Assert that CALL is to a function of the specified TYPE. It is
 ;;; assumed that the call is legal and has only constants in the
 ;;; keyword positions.
-(defun assert-call-type (call type &optional (trusted t))
+(defun assert-call-type (call type trusted where-from)
   (declare (type combination call) (type fun-type type))
   (let ((policy (lexenv-policy (node-lexenv call)))
         (returns (fun-type-returns type)))
@@ -1037,10 +1023,11 @@ and no value was provided for it." name))))))))))
           ;; less use there, but they can cause the MV call conversion
           ;; to cause astray.
           (when (and lvar
-                     (not (return-p (lvar-dest lvar)))
-                     (not (mv-combination-p (lvar-dest lvar)))
-                     (lvar-has-single-use-p lvar))
-            (when (assert-lvar-type lvar returns policy)
+                     (or (eq where-from :declared-verify)
+                         (and (not (return-p (lvar-dest lvar)))
+                              (not (mv-combination-p (lvar-dest lvar)))
+                              (lvar-has-single-use-p lvar))))
+            (when (assert-lvar-type lvar returns policy 'ftype-context)
               (reoptimize-lvar lvar)))))
     (let* ((name (lvar-fun-name (combination-fun call) t))
            (info (and name
@@ -1052,7 +1039,9 @@ and no value was provided for it." name))))))))))
            (lambda (arg type lvars &optional annotation)
              (when (and
                     (apply-type-annotation name arg type
-                                           lvars policy annotation)
+                                           lvars policy annotation
+                                           (and (not trusted)
+                                                'ftype-context))
                     (not trusted))
                (reoptimize-lvar arg)))
            call
