@@ -339,8 +339,7 @@
             (when (singleton-p (lvar-uses lvar))
               (barf "~S has exactly 1 use, but LVAR-USES is a list."
                     lvar))
-            (unless (or (lvar-dest lvar)
-                        (enclose-p node))
+            (unless (lvar-dest lvar)
               (barf "~S does not have dest." lvar))))
 
         (check-node-reached node)
@@ -503,6 +502,7 @@
        (let ((enclose (functional-enclose fun)))
          (unless (eq node enclose)
            (barf "~S is not the ENCLOSE for its FUN ~S." node enclose)))))
+    (cdynamic-extent)
     (exit
      (let ((entry (exit-entry node))
            (value (exit-value node)))
@@ -829,11 +829,11 @@
              (let ((tn (tn-ref-tn op))
                    (load-tn (tn-ref-load-tn op)))
                (unless (or (eq (tn-kind tn) :unused)
-                           (eq (svref (car scs)
-                                      (sc-number
-                                       (tn-sc
-                                        (or load-tn tn))))
-                               t))
+                           (typep (svref (car scs)
+                                         (sc-number
+                                          (tn-sc
+                                           (or load-tn tn))))
+                                  '(or (eql t) function)))
                  (barf "operand restriction not satisfied: ~S" op))))))
     (do-ir2-blocks (block component)
       (do ((vop (ir2-block-last-vop block) (vop-prev vop)))
@@ -927,11 +927,16 @@
      (format stream "~S {~A}" (leaf-debug-name leaf) (global-var-kind leaf)))
     (clambda
      (format stream "lambda ~@[~S ~]~:S"
-             (and (leaf-has-source-name-p leaf)
-                  (functional-debug-name leaf))
+             (if (leaf-has-source-name-p leaf)
+                 (leaf-source-name leaf)
+                 (functional-debug-name leaf))
              (mapcar #'leaf-debug-name (lambda-vars leaf))))
     (optional-dispatch
-     (format stream "optional-dispatch ~S" (mapcar #'leaf-debug-name (optional-dispatch-arglist leaf))))
+     (format stream "optional-dispatch ~a ~S"
+             (if (leaf-has-source-name-p leaf)
+                 (leaf-source-name leaf)
+                 (functional-debug-name leaf))
+             (mapcar #'leaf-debug-name (optional-dispatch-arglist leaf))))
     (functional
      (case (functional-kind leaf)
        (:toplevel-xep
@@ -973,7 +978,9 @@
 (defun print-lvar-stack (stack &optional (stream *standard-output*))
   (loop for (lvar . rest) on stack
         do (format stream "~:[u~;d~]v~D~@[ ~]"
-                   (lvar-dynamic-extent lvar) (cont-num lvar) rest)))
+                   (eq (ir2-lvar-kind (lvar-info lvar)) :stack)
+                   (cont-num lvar)
+                   rest)))
 
 (defvar *debug-print-types* nil)
 (defvar *debug-print-vop-temps* nil)
@@ -989,6 +996,10 @@
       (format t " <deleted>"))
 
     (pprint-newline :mandatory)
+    (let ((pred (block-pred block)))
+      (format t "predecessors~{ c~D~}~%"
+              (mapcar (lambda (x) (cont-num (block-start x))) pred)))
+    (pprint-newline :mandatory)
     (when (block-start-cleanup block)
       (format t "cleanup ~s~%" (cleanup-kind (block-start-cleanup block))))
     (awhen (block-info block)
@@ -1036,13 +1047,7 @@
            (print-lvar (return-result node))
            (print-leaf (return-lambda node)))
           (entry
-           (let ((cleanup (entry-cleanup node)))
-             (case (cleanup-kind cleanup)
-               ((:dynamic-extent)
-                (format t "entry DX~{ ~D~}"
-                        (cleanup-nlx-info cleanup)))
-               (t
-                (format t "entry ~S" (entry-exits node))))))
+           (format t "entry ~S" (entry-exits node)))
           (exit
            (let ((value (exit-value node)))
              (cond (value
@@ -1076,7 +1081,9 @@
                        (leaf (print-leaf thing))
                        (nlx-info (princ thing))))
                    (write-string "}"))))
-             (write-char #\space))))
+             (write-char #\space)))
+          (cdynamic-extent
+           (format t "dynamic extent ~S" (dynamic-extent-values node))))
         (when (and *debug-print-types*
                    (valued-node-p node))
           (write-char #\space)
@@ -1441,3 +1448,47 @@ is replaced with replacement."
                                   (cleanup-mess-up
                                    (block-start-cleanup ep)))))))))
     (write-line "}" stream)))
+
+(defun print-constraint (constraint &optional kind)
+  (let ((kind* (constraint-kind constraint)))
+    (when (or (not kind)
+              (eq kind* kind))
+      (flet ((f (x)
+               (cond ((lambda-var-p x)
+                      (leaf-source-name x))
+                     ((ctype-p x)
+                      (type-specifier x))
+                     (t x))))
+        (format t "   ~a~a ~a ~a~%" (if (eq kind* 'equality)
+                                        (format nil "~a ~a~@[ ~a~]" kind* (equality-constraint-operator constraint)
+                                                (when (/= (equality-constraint-amount constraint) 0)
+                                                  (equality-constraint-amount constraint)))
+                                        kind*)
+                (if (constraint-not-p constraint)
+                    " NOT"
+                    "")
+                (f (constraint-x constraint))
+                (f (constraint-y constraint)))))))
+
+(defun print-conset (conset &optional kind)
+  (do-conset-elements (con conset)
+    (print-constraint con kind)))
+
+(defun print-constraints (component &optional kind)
+  (do-blocks (block component)
+    (handler-case (progn
+                    (terpri)
+                    (terpri)
+                    ;; (do-conset-elements (con (block-in block))
+                    ;;   (print-constraint con kind))
+                    (print-nodes block)
+                    (let ((last (block-last block)))
+                      (cond ((if-p last)
+                             (format t "  CONSEQ~%")
+                             (print-conset (if-consequent-constraints last) kind)
+                             (format t "  ALT~%")
+                             (print-conset (if-alternative-constraints last) kind))
+                            (t
+                             (print-conset (block-out block) kind)))))
+      (error (condition)
+        (format t "~&~A...~%" condition)))))
