@@ -10,7 +10,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
+(in-package "SB-IMPL")
 
 ;;; Similar to FUNCTION, but the result type is "exactly" specified:
 ;;; if it is an object type, then the function returns exactly one
@@ -21,7 +21,7 @@
                       ((or (atom result)
                            (not (eq (car result) 'values)))
                        `(values ,result &optional))
-                      ((intersection (cdr result) sb!xc:lambda-list-keywords)
+                      ((intersection (cdr result) lambda-list-keywords)
                        result)
                       (t `(values ,@(cdr result) &optional)))))
     `(function ,args ,result)))
@@ -33,7 +33,50 @@
 ;;; (:FUNCTION :TYPE) information is extracted through a wrapper.
 ;;; The globaldb representation is not necessarily literally a CTYPE.
 #-sb-xc-host
-(declaim (ftype (sfunction (t) (values ctype boolean)) proclaimed-ftype))
+(declaim (ftype (sfunction (t) ctype) global-ftype))
+
+;;; A bit about the physical representation of the packed info format:
+;;; With #+compact-instance-header it is possible to represent a vector of N things
+;;; in a structure using (ALIGN-UP (1+ N) 2) words of memory. This is a saving
+;;; of 1 word on average when compared to SIMPLE-VECTOR which needs
+;;; (ALIGN-UP (+ N 2) 2) words.  Granted that either might have a padding word,
+;;; but I've observed 5% to 10% space reduction by eliminating one slot.
+;;; Without compact-instance-header, we'e indifferent, in terms of space,
+;;; as to whether this is an INSTANCE or a SIMPLE-VECTOR. For consistency,
+;;; we use an INSTANCE regardless of presence of the compact-header feature.
+;;; This makes assembly routines (e.g. CALL-SYMBOL) slightly less sensitive
+;;; to the feature's absence.
+
+;;; Since variable-length instances aren't portably a thing,
+;;; we use a structure of one slot holding a vector.
+#+sb-xc-host (defstruct (packed-info
+                          (:constructor %make-packed-info (cells))
+                          (:copier nil))
+               ;; These objects are immutable.
+               (cells #() :type simple-vector :read-only t))
+;;; Some abstractions for host/target compatibility of all the defuns.
+#+sb-xc-host
+(progn
+  (defmacro make-packed-info (n) `(%make-packed-info (make-array ,n)))
+  (defmacro copy-packed-info (info)
+    `(%make-packed-info (copy-seq (packed-info-cells ,info))))
+  (defmacro packed-info-len (info)
+    `(length (packed-info-cells ,info)))
+  (defmacro  %info-ref (info index) `(svref (packed-info-cells ,info) ,index)))
+#-sb-xc-host
+(progn
+  #+nil
+  (defmethod print-object ((obj packed-info) stream)
+    (format stream "[~{~W~^ ~}]"
+            (loop for i below (%instance-length obj)
+               collect (%instance-ref obj i))))
+  (defmacro packed-info-len (info)
+      `(- (%instance-length ,info) #.sb-vm:instance-data-start))
+  (defmacro %info-ref (v i)
+    `(%instance-ref ,v (+ ,i #.sb-vm:instance-data-start)))
+  (defmethod print-object ((self packed-info) stream)
+    (print-unreadable-object (self stream :type t :identity t)
+      (format stream "len=~d" (packed-info-len self)))))
 
 ;;; At run time, we represent the type of a piece of INFO in the globaldb
 ;;; by a small integer between 1 and 63.  [0 is reserved for internal use.]
@@ -41,14 +84,13 @@
 ;;; is not seen by the "#." expression a few lines down.
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (defconstant info-number-bits 6))
-(def!type info-number () `(unsigned-byte ,info-number-bits))
+(deftype info-number () `(unsigned-byte ,info-number-bits))
 
 ;;; A map from info-number to its META-INFO object.
 ;;; The reverse mapping is obtained by reading the META-INFO.
 (declaim (type (simple-vector #.(ash 1 info-number-bits)) *info-types*))
-(!define-load-time-global *info-types*
-            ;; Must be dumped as a literal for cold-load.
-            #.(make-array (ash 1 info-number-bits) :initial-element nil))
+(define-load-time-global *info-types* (make-array (ash 1 info-number-bits) :initial-element nil))
+#+sb-xc-host (defvar *get-info-value-histo* (make-array 64))
 
 (defstruct (meta-info
             (:constructor
@@ -81,51 +123,43 @@
 (defconstant +info-metainfo-type-num+ 0)
 
 ;; Refer to info-vector.lisp for the meaning of this constant.
-(defconstant +no-auxilliary-key+ 0)
+(defconstant +no-auxiliary-key+ 0)
 
-;; Return the globaldb info for SYMBOL. With respect to the state diagram
-;; presented at the definition of SYMBOL-PLIST, if the object in SYMBOL's
-;; info slot is LISTP, it is in state 1 or 3. Either way, take the CDR.
-;; Otherwise, it is in state 2 so return the value as-is.
-;; In terms of this function being named "-vector", implying always a vector,
-;; it is understood that NIL is a proxy for +NIL-PACKED-INFOS+, a vector.
-;;
+;;; Return the globaldb info for SYMBOL. With respect to the state diagram
+;;; presented at the definition of SYMBOL-PLIST, if the object in SYMBOL's
+;;; info slot is LISTP, it is in state 1 or 3. Either way, take the CDR.
+;;; Otherwise, it is in state 2 so return the value as-is.
+;;; NIL is an acceptable substitute for +NIL-PACKED-INFOS+,
+;;; but I might change that.
 #-sb-xc-host
 (progn
- ;; Don't inline this if a vop translates it. Inlining occurs first,
- ;; causing the vop not to be used.
- #!-(vop-translates sb!kernel:symbol-info-vector) (declaim (inline symbol-info-vector))
- (defun symbol-info-vector (symbol)
-  (let ((info-holder (symbol-info symbol)))
-    (truly-the (or null simple-vector)
+(declaim (inline symbol-dbinfo))
+(defun symbol-dbinfo (symbol)
+  (let ((info-holder (symbol-%info symbol)))
+    (truly-the (or null packed-info)
                (if (listp info-holder) (cdr info-holder) info-holder)))))
 
-;;; SYMBOL-INFO is a primitive object accessor defined in 'objdef.lisp'
-;;; But in the host Lisp, there is no such thing as a symbol-info slot.
-;;; Instead, symbol-info is kept in the host symbol's plist.
-;;; This must be a SETFable place.
-#+sb-xc-host
-(defmacro symbol-info-vector (symbol) `(get ,symbol :sb-xc-globaldb-info))
-
 ;; Perform the equivalent of (GET-INFO-VALUE KIND +INFO-METAINFO-TYPE-NUM+)
 ;; but skipping the defaulting logic.
 ;; Return zero or more META-INFOs that match on KIND, which is usually
 ;; - though not always - a unique identifier for the (:TYPE :KIND) pair.
 ;; Note that bypassing of defaults is critical for bootstrapping,
 ;; since INFO is used to retrieve its own META-INFO at system-build time.
-(defmacro !get-meta-infos (kind)
-  `(let* ((info-vector (symbol-info-vector ,kind))
-          (index (if info-vector
-                     (packed-info-value-index info-vector +no-auxilliary-key+
+(defmacro get-meta-infos (kind)
+  `(let* ((packed-info (symbol-dbinfo ,kind))
+          (index (if packed-info
+                     (packed-info-value-index packed-info +no-auxiliary-key+
                                               +info-metainfo-type-num+))))
-     (if index (svref info-vector index))))
+     (if index (%info-ref packed-info index))))
 
-;; (UNSIGNED-BYTE 16) is an arbitrarily generous limit on the number of
-;; cells in an info-vector. Most vectors have a fewer than a handful of things,
+;; (UNSIGNED-BYTE 11) is an arbitrarily generous limit on the number of
+;; cells in a packed-info. Most packed-infos have fewer than a handful of things,
 ;; and performance would need to be re-thought if more than about a dozen
 ;; cells were in use. (It would want to become hash-based probably)
-(declaim (ftype (function (simple-vector (or (eql 0) symbol) info-number)
-                          (or null (unsigned-byte 16)))
+;; It has to be smaller than INSTANCE_LENGTH_MASK certainly,
+;; plus leaving room for a layout slot if #-compact-instance-header.
+(declaim (ftype (function (packed-info (or (eql 0) symbol) info-number)
+                          (or null (unsigned-byte 11)))
                 packed-info-value-index))
 
 ;; Return the META-INFO object for CATEGORY and KIND, signaling an error
@@ -138,7 +172,7 @@
 ;; through, whereas typically no more than 3 or 4 items have the same KIND.
 ;;
 (defun meta-info (category kind &optional (errorp t))
-  (or (let ((metadata (!get-meta-infos kind)))
+  (or (let ((metadata (get-meta-infos kind)))
         (cond ((listp metadata) ; conveniently handles NIL
                (dolist (info metadata nil) ; FIND is slower :-(
                  (when (eq (meta-info-category (truly-the meta-info info))
@@ -146,7 +180,7 @@
                    (return info))))
               ((eq (meta-info-category (truly-the meta-info metadata)) category)
                metadata)))
-      ;; !GET-META-INFOS enforces that KIND is a symbol, therefore
+      ;; GET-META-INFOS enforces that KIND is a symbol, therefore
       ;; if a metaobject was found, CATEGORY was necessarily a symbol too.
       ;; Otherwise, if the caller wants no error to be signaled on missing info,
       ;; we must nevertheless enforce that CATEGORY was actually a symbol.
@@ -163,33 +197,34 @@
 ;;; and when running target code, calls with legal constants for the first two
 ;;; arguments are transformed.
 (macrolet ((def (name lambda-list form)
+             ;; AVER is not defined yet
              (assert (and (member 'category lambda-list)
                           (member 'kind lambda-list)))
              `(define-compiler-macro ,name ,(append '(&whole .whole.) lambda-list)
-                (if (and (keywordp category) (keywordp kind))
-                    ;; In the target Lisp, it's a STYLE-WARNING if this macro
-                    ;; defers to a full call to #'INFO.
-                    ;; If the cross-compilation host, if any info-type is
-                    ;; defined, then it's an error not to find the meta-info.
-                    ;; If no info-types are defined, silently defer.
-                    (let ((meta-info
-                           (and #+sb-xc-host (find-if #'identity *info-types*)
-                                (meta-info category kind #-sb-xc-host nil))))
-                      (if meta-info
-                          ,form
-                          (progn
-                            #-sb-xc-host
-                            (style-warn "(~S ~S) is not a defined info type."
-                                        category kind)
-                            .whole.)))
-                    .whole.))))
+                (when (and (keywordp category) (keywordp kind))
+                  ;; In the target Lisp, it's a STYLE-WARNING if this macro
+                  ;; defers to a full call to #'INFO.
+                  ;; If the cross-compilation host, if any info-type is
+                  ;; defined, then it's an error not to find the meta-info.
+                  ;; If no info-types are defined, silently defer.
+                  (let ((meta-info (meta-info category kind #+sb-xc-host nil)))
+                    (when meta-info
+                      (return-from ,(if (eq name 'clear-info) 'clear-info 'info)
+                        ,form))))
+                ;; This flags some calls in make-host-2 that have non-constant args
+                ;; (format t "~&Could not xform ~S ~S ~S~%" ',name category kind)
+                .whole.)))
 
+  ;; ECL bug workaround: INFO ceases to be a valid macrolet name
+  ;; because it tries to run the compiler-macro before the local macro.
+  ;; In particular, "(collect ((info)) ...)" will not compile correctly.
+  #-host-quirks-ecl
   (def info (category kind name)
     `(truly-the (values ,(meta-info-type-spec meta-info) boolean)
                 (get-info-value ,name ,(meta-info-number meta-info))))
 
   (def (setf info) (new-value category kind name)
-    (let* (#+sb-xc-host (sb!xc:*gensym-counter* sb!xc:*gensym-counter*)
+    (let* (#+sb-xc-host (sb-xc:*gensym-counter* sb-xc:*gensym-counter*)
            (tin (meta-info-number meta-info)) ; info-type id number
            (type-spec (meta-info-type-spec meta-info))
            (new (make-symbol "NEW"))
@@ -217,18 +252,12 @@
 ;; Atomic update will be important for making the fasloader threadsafe
 ;; using a predominantly lock-free design, and other nice things.
 (defmacro atomic-set-info-value (category kind name lambda)
-  (with-unique-names (info-number proc)
+  (with-unique-names (info-number)
     `(let ((,info-number
             ,(if (and (keywordp category) (keywordp kind))
                  (meta-info-number (meta-info category kind))
                  `(meta-info-number (meta-info ,category ,kind)))))
-       ,(if (and (listp lambda) (eq (car lambda) 'lambda))
-            ;; rewrite as FLET because the compiler is unable to dxify
-            ;;   (DX-LET ((x (LAMBDA <whatever>))) (F x))
-            (destructuring-bind (lambda-list . body) (cdr lambda)
-              `(dx-flet ((,proc ,lambda-list ,@body))
-                 (%atomic-set-info-value ,name ,info-number #',proc)))
-            `(%atomic-set-info-value ,name ,info-number ,lambda)))))
+       (%atomic-set-info-value ,name ,info-number ,lambda))))
 
 ;; Perform the approximate equivalent operations of retrieving
 ;; (INFO :CATEGORY :KIND NAME), but if no info is found, invoke CREATION-FORM
@@ -241,104 +270,58 @@
 ;; A mutex-guarded table would probably be more appropriate in such cases.
 ;;
 (defmacro get-info-value-initializing (category kind name creation-form)
-  (let ((proc (make-symbol "THUNK")))
-    `(dx-flet ((,proc () ,creation-form))
-       (%get-info-value-initializing
-        ,(if (and (keywordp category) (keywordp kind))
-             (meta-info-number (meta-info category kind))
-             `(meta-info-number (meta-info ,category ,kind)))
-        ,name #',proc))))
+  `(%get-info-value-initializing
+    ,(if (and (keywordp category) (keywordp kind))
+         (meta-info-number (meta-info category kind))
+         `(meta-info-number (meta-info ,category ,kind)))
+    ,name (lambda () ,creation-form)))
 
-;;;; boolean attribute utilities
-;;;;
-;;;; We need to maintain various sets of boolean attributes for known
-;;;; functions and VOPs. To save space and allow for quick set
-;;;; operations, we represent the attributes as bits in a fixnum.
+#+sb-xc-host
+(progn
+(defun calc-globaldb-lookup-freq-dist ()
+  (sort (loop with histo = *get-info-value-histo*
+              for i below 64
+              when (plusp (aref histo i))
+              collect (let ((info (aref *info-types* i)))
+                        (cons (cons (meta-info-category info) (meta-info-kind info))
+                              (aref histo i))))
+        #'> :key #'cdr))
+(defun print-globaldb-lookup-freq-dist (list)
+  (dolist (x list) (format t "~8d ~s~%" (cdr x) (car x)))))
 
-(in-package "SB!C")
-(deftype attributes () 'fixnum)
-
-;;; Given a list of attribute names and an alist that translates them
-;;; to masks, return the OR of the masks.
-(defun encode-attribute-mask (names universe)
-  (loop for name in names
-        for pos = (position name universe)
-        sum (if pos (ash 1 pos) (error "unknown attribute name: ~S" name))))
-
-(defun decode-attribute-mask (bits universe)
-  (loop for name across universe
-        for mask = 1 then (ash mask 1)
-        when (logtest mask bits) collect name))
-
-;;; Define a new class of boolean attributes, with the attributes
-;;; having the specified ATTRIBUTE-NAMES. NAME is the name of the
-;;; class, which is used to generate some macros to manipulate sets of
-;;; the attributes:
-;;;
-;;;    NAME-attributep attributes attribute-name*
-;;;      Return true if one of the named attributes is present, false
-;;;      otherwise. When set with SETF, updates the place Attributes
-;;;      setting or clearing the specified attributes.
-;;;
-;;;    NAME-attributes attribute-name*
-;;;      Return a set of the named attributes.
-(defmacro !def-boolean-attribute (name &rest attribute-names)
-  (let ((vector (coerce attribute-names 'vector))
-        (constructor (symbolicate name "-ATTRIBUTES"))
-        (test-name (symbolicate name "-ATTRIBUTEP")))
-    `(progn
-       (defmacro ,constructor (&rest attribute-names)
-         "Automagically generated boolean attribute creation function.
-  See !DEF-BOOLEAN-ATTRIBUTE."
-         (encode-attribute-mask attribute-names ,vector))
-       (defun ,(symbolicate "DECODE-" name "-ATTRIBUTES") (attributes)
-         (decode-attribute-mask attributes ,vector))
-       (defmacro ,test-name (attributes &rest attribute-names)
-         "Automagically generated boolean attribute test function.
-  See !DEF-BOOLEAN-ATTRIBUTE."
-         `(logtest (the attributes ,attributes)
-                   (,',constructor ,@attribute-names)))
-       (define-setf-expander ,test-name (place &rest attributes
-                                               &environment env)
-         "Automagically generated boolean attribute setter. See
- !DEF-BOOLEAN-ATTRIBUTE."
-         (multiple-value-bind (temps values stores setter getter)
-             (#+sb-xc-host get-setf-expansion
-              #-sb-xc-host sb!xc:get-setf-expansion place env)
-           (when (cdr stores)
-             (error "multiple store variables for ~S" place))
-           (let ((newval (sb!xc:gensym))
-                 (n-place (sb!xc:gensym))
-                 (mask (encode-attribute-mask attributes ,vector)))
-             (values `(,@temps ,n-place)
-                     `(,@values ,getter)
-                     `(,newval)
-                     `(let ((,(first stores)
-                             (if ,newval
-                                 (logior ,n-place ,mask)
-                                 (logandc2 ,n-place ,mask))))
-                        ,setter
-                        ,newval)
-                     `(,',test-name ,n-place ,@attributes))))))))
-
-;;; And now for some gratuitous pseudo-abstraction...
-;;;
-;;; ATTRIBUTES-UNION
-;;;   Return the union of all the sets of boolean attributes which are its
-;;;   arguments.
-;;; ATTRIBUTES-INTERSECTION
-;;;   Return the intersection of all the sets of boolean attributes which
-;;;   are its arguments.
-;;; ATTRIBUTES=
-;;;   True if the attributes present in ATTR1 are identical to
-;;;   those in ATTR2.
-(defmacro attributes-union (&rest attributes)
-  `(the attributes
-        (logior ,@(mapcar (lambda (x) `(the attributes ,x)) attributes))))
-(defmacro attributes-intersection (&rest attributes)
-  `(the attributes
-        (logand ,@(mapcar (lambda (x) `(the attributes ,x)) attributes))))
-(declaim (ftype (function (attributes attributes) boolean) attributes=))
-#!-sb-fluid (declaim (inline attributes=))
-(defun attributes= (attr1 attr2)
-  (eql attr1 attr2))
+;;; This priority order could be computed during make-host-2 based on what calls
+;;; we observe, and then injected into the target compiler. But instead I'm just
+;;; hardcoding what I observed on a particular run of make-host-2 based on
+;;; the output of PRINT-GLOBALDB-LOOKUP-FREQ-DIST. It won't change much.
+(defglobal *info-priority-order*
+    '(nil ; 0 can't be used
+      (:type . :builtin)
+      (:function . :info)
+      (:function . :ir1-convert)
+      (:function . :source-transform)
+      (:function . :kind)
+      (:variable . :kind)
+      (:type . :classoid-cell)
+      (:function . :compiler-macro-function)
+      (:function . :deprecated)
+      (:function . :inlinep)
+      (:function . :type)
+      (:function . :macro-function)
+      (:function . :where-from)
+      (:type . :kind)
+      (:function . :inlining-data)
+      (:function . :assumed-type)
+      (:type . :deprecated)
+      (:type . :expander)
+      (:setf . :expander)
+      (:type . :compiler-layout)
+      (:variable . :wired-tls)
+      (:function . :walker-template)
+      (:variable . :deprecated)
+      (:variable . :macro-expansion)
+      (:variable . :where-from)
+      (:variable . :type)
+      (:variable . :always-bound)
+      (:source-location . :declaration)
+      (:alien-type . :translator)
+      (:alien-type . :kind)))
