@@ -55,9 +55,10 @@
                             (type-specifier
                              (specifier-type
                               `(simple-array ,(sb-vm:saetp-specifier saetp) (*)))))
-                          (remove t sb-vm:*specialized-array-element-type-properties*
-                                  :key 'sb-vm:saetp-specifier))))
-                `(((integer 0 ,sb-xc:array-dimension-limit)
+                          (remove-if (lambda (x) (member x '(nil t)))
+                                     sb-vm:*specialized-array-element-type-properties*
+                                     :key 'sb-vm:saetp-specifier))))
+                `(((mod ,(1+ array-dimension-limit))
                    object-not-array-dimension)
                   ;; Union of all unboxed array specializations,
                   ;; for type-checking the argument to VECTOR-SAP
@@ -72,16 +73,20 @@
         ;; Error number must be of type (unsigned-byte 8).
         (assert (<= (length list) 256))
         `(defconstant-eqx sb-c:+backend-internal-errors+
-               ,(map 'vector
-                     (lambda (x)
+             ,(map 'vector
+                   (lambda (x)
+                     (flet ((normalize-type (type)
+                              (if (stringp type)
+                                  type
+                                  (type-specifier (specifier-type type)))))
                        (if (symbolp x)
-                           (list* x (symbolicate "OBJECT-NOT-" x "-ERROR") 1)
-                           (list* (car x) (symbolicate (second x) "-ERROR")
+                           (list* (normalize-type x) (symbolicate "OBJECT-NOT-" x "-ERROR") 1)
+                           (list* (normalize-type (car x)) (symbolicate (second x) "-ERROR")
                                   (if (stringp (car x))
                                       (third x)
-                                      1))))
-                     list)
-               #'equalp))))
+                                      1)))))
+                   list)
+           #'equalp))))
  (compute-it
   ;; Keep the following two subsets of internal errors in this order:
   ;;
@@ -93,23 +98,41 @@
    ("An attempt was made to use an undefined FDEFINITION." undefined-fun 1)
    #+(or arm arm64 x86-64)
    ("An attempt was made to use an undefined alien function" undefined-alien-fun 1)
+   ;; Only x86-64 can detect uninitialized C memory
+   ;; but there's no harm in defining this error.
+   ("Read of uninitialized memory" uninitialized-memory 0)
    ("invalid argument count" invalid-arg-count 1)
    ("invalid argument count" local-invalid-arg-count 2)
    ("bogus argument to VALUES-LIST" bogus-arg-to-values-list 1)
    ("An attempt was made to use an undefined SYMBOL-VALUE." unbound-symbol 1)
    ("attempt to RETURN-FROM a block that no longer exists" invalid-unwind 0)
    ("attempt to THROW to a non-existent tag" unseen-throw-tag 1)
-   ("division by zero" division-by-zero 2)
+   ("division by zero" division-by-zero 1)
    ("Object is of the wrong type." object-not-type 2)
+   ("ECASE failure" ecase-failure 2)
+   ("ETYPECASE failure" etypecase-failure 2)
    ("odd number of &KEY arguments" odd-key-args 0)
    ("unknown &KEY argument" unknown-key-arg 1)
    ("invalid array index" invalid-array-index 3)
+   ("invalid vector index" invalid-vector-index 2)
+   ("uninitialized element" uninitialized-element 2)
    ("A function with declared result type NIL returned." nil-fun-returned 1)
    ("An array with element-type NIL was accessed." nil-array-accessed 1)
    ("Object layout is invalid. (indicates obsolete instance)" layout-invalid 2)
    ("Thread local storage exhausted." tls-exhausted 0)
-   ("Failed aver" failed-aver 1))
-
+   ("Stack allocated object overflows stack." stack-allocated-object-overflows-stack 1)
+   ("Unreachable code reached" unreachable 0)
+   ("Failed aver" failed-aver 1)
+   ("Multiplication overflow" mul-overflow 2)
+   ("Add overflow" add-sub-overflow 1)
+   #+x86-64
+   ("Sub overflow" sub-overflow 1)
+   ("Add overflow" signed-unsigned-add-overflow 1)
+   ("Add overflow" add-overflow2 2)
+   ("Sub overflow" sub-overflow2 2)
+   ("Mul overflow" mul-overflow2 2)
+   ("ASH overflow" ash-overflow2 2)
+   ("Negate overflow" negate-overflow 1))
   ;; (II) All the type specifiers X for which there is a unique internal
   ;;      error code corresponding to a primitive object-not-X-error.
   function
@@ -214,7 +237,16 @@
   sb-format::format-directive
   package
   form-tracking-stream
-  ansi-stream))
+  ansi-stream
+  ((unsigned-byte 16) object-not-unsigned-byte-16)
+  ((signed-byte 8) object-not-signed-byte-8)
+  ((signed-byte 16) object-not-signed-byte-16)
+  ((or index list) object-not-index-or-list)
+  condition
+  sb-pcl::fast-method-call
+  ((or symbol string) object-not-or-symbol-string)
+  ((and unsigned-byte fixnum) object-not-unsigned-fixnum)
+  bit-index))
 
 (defun error-number-or-lose (name)
   (or (position name sb-c:+backend-internal-errors+
@@ -225,20 +257,3 @@
   (if (array-in-bounds-p sb-c:+backend-internal-errors+ error-number)
       (cddr (svref sb-c:+backend-internal-errors+ error-number))
       0))
-
-#-sb-xc-host ; no SB-C:SAP-READ-VAR-INTEGERF
-(defun decode-internal-error-args (sap trap-number &optional error-number)
-  (let ((error-number (cond (error-number)
-                            ((>= trap-number sb-vm:error-trap)
-                             (prog1
-                                 (- trap-number sb-vm:error-trap)
-                               (setf trap-number sb-vm:error-trap)))
-                            (t
-                             (prog1 (sap-ref-8 sap 0)
-                               (setf sap (sap+ sap 1)))))))
-    (let ((length (sb-kernel::error-length error-number)))
-      (declare (type (unsigned-byte 8) length))
-      (values error-number
-              (loop repeat length with index = 0
-                    collect (sb-c:sap-read-var-integerf sap index))
-              trap-number))))
