@@ -76,7 +76,9 @@
     (and static-fun-index
          (+ (* (length +static-symbols+) (pad-data-block symbol-size))
             (pad-data-block (1- symbol-size))
-            (* 4 n-word-bytes) ; sizeof SB-LOCKLESS:+TAIL+
+            ;; sizeof SB-LOCKLESS:+TAIL+ is calculated as 1 user data slot,
+            ;; round-to-odd, add the header word.
+            (* (1+ (logior (1+ sb-vm:instance-data-start) 1)) n-word-bytes)
             (- list-pointer-lowtag)
             (* static-fun-index (pad-data-block fdefn-size))
             other-pointer-lowtag))))
@@ -208,15 +210,6 @@
   "Cause a continuable error.  ERROR-CODE is the error to cause."
   (emit-error-break vop cerror-trap (error-number-or-lose error-code) values))
 
-;; C header generation can't execute DEFINE-VOP
-;; Perhaps this belongs somewhere different
-#+(and sb-safepoint (not :c-headers-only))
-(define-vop (insert-safepoint)
-  (:policy :fast-safe)
-  (:translate sb-kernel::gc-safepoint)
-  (:generator 0
-    (emit-safepoint)))
-
 ;;; Does the TN definitely hold *any* of the 4 pointer types
 (defun pointer-tn-ref-p (tn-ref)
   (and (sc-is (tn-ref-tn tn-ref) descriptor-reg)
@@ -354,7 +347,7 @@
 (define-load-time-global *store-barriers-potentially-emitted* 0)
 (define-load-time-global *store-barriers-emitted* 0)
 
-(defun require-gengc-barrier-p (object value-tn-ref value-tn)
+(defun require-gengc-barrier-p (object value-tn-ref value-tn &optional allocator)
   (incf *store-barriers-potentially-emitted*)
   ;; If OBJECT is stack-allocated, elide the barrier
   (when (stack-consed-p object)
@@ -368,17 +361,33 @@
            ;; If immediate non-pointer, elide the barrier
            (when (sc-is tn immediate)
              (let ((value (tn-value tn)))
-               (when (sb-xc:typep value '(or character sb-xc:fixnum boolean
-                                             #+64-bit single-float))
+               (when (sb-xc:typep value '(or character sb-xc:fixnum
+                                          #+64-bit single-float
+                                          boolean))
                  (return-from potential-heap-pointer-p nil))))
            (when (sb-c::unbound-marker-tn-p tn)
              (return-from potential-heap-pointer-p nil))
            ;; And elide for things like (OR FIXNUM NULL)
            (let ((type (tn-ref-type tn-ref)))
-             (when (csubtypep type (specifier-type '(or character sb-xc:fixnum boolean
-                                                        #+64-bit single-float)))
+             (when (or (csubtypep type #1=(specifier-type '(or character sb-xc:fixnum boolean
+                                                         #+64-bit single-float)))
+                       (let ((diff (type-difference type #1#)))
+                         (and (member-type-p diff)
+                              #-sb-xc-host
+                              (loop for member in (member-type-members diff)
+                                    always
+                                    (and (eql (generation-of member) +pseudo-static-generation+)
+                                         (or (not (sb-c::producing-fasl-file))
+                                             (and (symbolp member)
+                                                  (logtest +symbol-initial-core+ (get-header-data member)))))))))
                (return-from potential-heap-pointer-p nil)))
-           t))
+           t)
+         (boxed-tn-p (value-tn)
+           (let* ((prim-type (sb-c::tn-primitive-type value-tn))
+                  (scs (and prim-type
+                            (sb-c::primitive-type-scs prim-type))))
+             (or (singleton-p scs)
+                 (not (member descriptor-reg-sc-number scs))))))
     (cond (value-tn
            (unless (eq (tn-ref-tn value-tn-ref) value-tn)
              (aver (eq (tn-ref-load-tn value-tn-ref) value-tn)))
@@ -392,14 +401,35 @@
                     (when (potential-heap-pointer-p (tn-ref-tn ref) ref)
                       (return t)))))
              (unless any-pointer
-               (return-from require-gengc-barrier-p nil))))))
-  (incf *store-barriers-emitted*)
-  t)
+               (return-from require-gengc-barrier-p nil)))))
+    (let (why)
+      (cond ((and value-tn
+                  value-tn-ref
+                  ;; Can this TN be boxed after the allocator?
+                  (boxed-tn-p value-tn)
+                  (or (eq allocator :allocator)
+                      (setf why
+                            (sb-c::set-slot-old-p (sb-c::vop-node (tn-ref-vop value-tn-ref))
+                                                  (vop-arg-position value-tn-ref (tn-ref-vop value-tn-ref))))))
+             (values nil why))
+            (t
+             (incf *store-barriers-emitted*)
+             t)))))
 
 (defun vop-nth-arg (n vop)
   (let ((ref (vop-args vop)))
     (dotimes (i n ref) (setq ref (tn-ref-across ref)))))
 
+(defun vop-arg-position (tn-ref vop)
+  (let ((types (sb-c::vop-info-arg-types (sb-c::vop-info vop))))
+    (loop with i = -1
+          for ref = (vop-args vop) then (tn-ref-across ref)
+          do
+          (loop do (pop types)
+                   (incf i)
+                while (typep (car types) '(cons (eql :constant))))
+          when (eq ref tn-ref) return i)))
+
 (defun length-field-shift (widetag)
   (if (= widetag instance-widetag)
       instance-length-shift
@@ -424,7 +454,7 @@
                 (let ((rank (- nwords array-dimensions-offset)))
                   (ash (encode-array-rank rank) array-rank-position))
                 (case widetag
-                  (#.fdefn-widetag 0)
+                  ((#.symbol-widetag #.fdefn-widetag) 0)
                   (t (ash (1- nwords) (length-field-shift widetag)))))
             widetag)))
 
@@ -518,3 +548,18 @@
         while e
         thereis (sb-c::lexenv-find 'sb-vm::.pseudo-atomic-call-out.
                                    vars :lexenv e)))
+
+;;; The SET vop should alway get a symbol that is known at compile-time
+;;; even though the compiler might have already done you a "favor" of
+;;; producing a load-TN for a random constant. This tries to undo that.
+(defun known-symbol-use-p (vop symbol)
+  (cond ((sc-is symbol constant immediate)
+         (tn-value symbol))
+        (t
+         ;; Given a DESCRIPTOR-REG you can refer back to the vop args
+         ;; to figure out what was loaded.
+         (let ((type (tn-ref-type (vop-args vop))))
+           ;; I'm pretty sure this _must_ be a MEMBER type.
+           (when (and (member-type-p type)
+                      (= (member-type-size type) 1))
+             (the symbol (first (member-type-members type))))))))
