@@ -12,7 +12,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 (/show0 "knownfun.lisp 17")
 
@@ -30,9 +30,9 @@
   (type (missing-arg) :type ctype)
   ;; the transformation function. Takes the COMBINATION node and
   ;; returns a lambda expression, or throws out.
-  (function (missing-arg) :type function)
-  ;; string used in efficiency notes
-  (note (missing-arg) :type string)
+  ;; If a cons, then the CAR is the function to call, and the CDR is an argument
+  ;; to pass to that function in addition to the NODE being considered.
+  (%fun (missing-arg) :type (or function (cons function)))
   ;; T if we should emit a failure note even if SPEED=INHIBIT-WARNINGS.
   (important nil :type (member nil :slightly t))
   ;; A function with NODE as an argument that checks wheteher the
@@ -41,29 +41,36 @@
   ;; notes about failed transformation due to types even though it
   ;; wouldn't have been applied with the right types anyway,
   ;; or if another transform could be applied with the right policy.
-  (policy nil :type (or null function))
-  (extra-type nil))
+  (policy nil :type (or null function)))
+(defun transform-function (transform)
+  (let ((fun (transform-%fun transform))) (if (listp fun) (car fun) fun)))
+(defun transform-note (transform)
+  (or #+sb-xc-host (documentation (transform-function transform) 'function)
+      #-sb-xc-host (and (fboundp 'sb-pcl::fun-doc)
+                        (funcall 'sb-pcl::fun-doc (transform-function transform)))
+      "optimize"))
 
-(defprinter (transform) type note important)
+(defmethod print-object ((x transform) stream)
+  (print-unreadable-object (x stream :type t :identity t)
+    (princ (type-specifier (transform-type x)) stream)))
 
 ;;; Grab the FUN-INFO and enter the function, replacing any old
 ;;; one with the same type and note.
-(defun %deftransform (name type fun &optional note important policy)
+;;; Argument order is: policy constraint, ftype constraint, consequent.
+;;; (think "qualifiers + specializers -> method")
+(defun %deftransform (name policy type fun &optional (important :slightly))
   (let* ((ctype (specifier-type type))
-         (note (or note "optimize"))
          (info (fun-info-or-lose name))
          (old (find ctype (fun-info-transforms info)
                     :test #'type=
                     :key #'transform-type)))
     (cond (old
-           (style-warn 'redefinition-with-deftransform
-                       :transform old)
-           (setf (transform-function old) fun
-                 (transform-note old) note
+           (style-warn 'redefinition-with-deftransform :transform old)
+           (setf (transform-%fun old) fun
                  (transform-important old) important
                  (transform-policy old) policy))
           (t
-           (push (make-transform :type ctype :function fun :note note
+           (push (make-transform :type ctype :%fun fun
                                  :important important
                                  :policy policy)
                  (fun-info-transforms info))))
@@ -75,38 +82,63 @@
                   &key derive-type optimizer result-arg
                        overwrite-fndb-silently
                        call-type-deriver
-                       annotation)
-  (let ((ctype (specifier-type type)))
+                       annotation
+                       folder)
+  (let* ((ctype (specifier-type type))
+         (type-to-store (if (contains-unknown-type-p ctype)
+                            ;; unparse it, so SFUNCTION -> FUNCTION
+                            (type-specifier ctype)
+                            ctype)))
     (dolist (name names)
-      (unless overwrite-fndb-silently
-        (let ((old-fun-info (info :function :info name)))
-          (when old-fun-info
-            ;; This is handled as an error because it's generally a bad
-            ;; thing to blow away all the old optimization stuff. It's
-            ;; also a potential source of sneaky bugs:
-            ;;    DEFKNOWN FOO
-            ;;    DEFTRANSFORM FOO
-            ;;    DEFKNOWN FOO ; possibly hidden inside some macroexpansion
-            ;;    ; Now the DEFTRANSFORM doesn't exist in the target Lisp.
-            ;; However, it's continuable because it might be useful to do
-            ;; it when testing new optimization stuff interactively.
-            (cerror "Go ahead, overwrite it."
-                    "~@<overwriting old FUN-INFO ~2I~_~S ~I~_for ~S~:>"
-                    old-fun-info name))))
-      (setf (info :function :type name) ctype)
-      (setf (info :function :where-from name) :declared)
-      (setf (info :function :kind name) :function)
-      (setf (info :function :info name)
-            (make-fun-info :attributes attributes
-                           :derive-type derive-type
-                           :optimizer optimizer
-                           :result-arg result-arg
-                           :call-type-deriver call-type-deriver
-                           :annotation annotation))
-      (if location
-          (setf (getf (info :source-location :declaration name) 'defknown)
-                location)
-          (remf (info :source-location :declaration name) 'defknown))))
+      (let ((old-fun-info (info :function :info name))
+            inherit)
+        (block ignore
+          (unless overwrite-fndb-silently
+            (when old-fun-info
+              ;; This is handled as an error because it's generally a bad
+              ;; thing to blow away all the old optimization stuff. It's
+              ;; also a potential source of sneaky bugs:
+              ;;    DEFKNOWN FOO
+              ;;    DEFTRANSFORM FOO
+              ;;    DEFKNOWN FOO ; possibly hidden inside some macroexpansion
+              ;;    ; Now the DEFTRANSFORM doesn't exist in the target Lisp.
+              ;; However, it's continuable because it might be useful to do
+              ;; it when testing new optimization stuff interactively.
+              (restart-case
+                  (cerror "Go ahead, overwrite it."
+                          "~@<overwriting old FUN-INFO ~2I~_~S ~I~_for ~S~:>"
+                          old-fun-info name)
+                (continue ()
+                  :report "Inherit templates and optimizers"
+                  (setf inherit t))
+                (ignore ()
+                  (return-from ignore)))))
+          (setf (info :function :type name) type-to-store)
+          (setf (info :function :where-from name) :declared)
+          (setf (info :function :kind name) :function)
+          (cond (inherit
+                 (when optimizer
+                   (setf (fun-info-optimizer old-fun-info) optimizer))
+                 (when derive-type
+                   (setf (fun-info-derive-type old-fun-info) derive-type))
+                 (setf (fun-info-attributes old-fun-info) attributes
+                       (fun-info-result-arg old-fun-info) result-arg
+                       (fun-info-annotation old-fun-info) annotation
+                       (fun-info-call-type-deriver old-fun-info) call-type-deriver
+                       (fun-info-folder old-fun-info) folder))
+                (t
+                 (setf (info :function :info name)
+                       (make-fun-info :attributes attributes
+                                      :derive-type derive-type
+                                      :optimizer optimizer
+                                      :result-arg result-arg
+                                      :call-type-deriver call-type-deriver
+                                      :annotation annotation
+                                      :folder folder))))
+          (if location
+              (setf (getf (info :source-location :declaration name) 'defknown)
+                    location)
+              (remf (info :source-location :declaration name) 'defknown))))))
   names)
 
 
@@ -142,6 +174,17 @@
     (setq attributes (union '(unwind) attributes)))
   (when (member 'flushable attributes)
     (pushnew 'unsafely-flushable attributes))
+  ;; Needs to be supported by the call VOPs
+  #-(or arm64 x86-64)
+  (setf attributes (remove 'no-verify-arg-count attributes))
+  #-(or arm64 x86-64)
+  (setf attributes (remove 'unboxed-return attributes))
+  #-(or arm64 x86-64) ;; Needs to be supported by the call VOPs, sb-vm::fixed-call-arg-location
+  (setf attributes (remove 'fixed-args attributes))
+  (when (or (memq 'fixed-args attributes)
+            (memq 'unboxed-return attributes))
+    (pushnew 'no-verify-arg-count attributes))
+
   (multiple-value-bind (type annotation)
       (split-type-info arg-types result-type)
     `(%defknown ',(if (and (consp name)
@@ -152,7 +195,14 @@
                 (ir1-attributes ,@attributes)
                 (source-location)
                 :annotation ,annotation
-                ,@keys)))
+                ,@keys
+                :folder ,(and (memq 'foldable attributes)
+                              (not (getf keys :folder))
+                              (or
+                               (memq 'fixed-args attributes)
+                               (memq 'unboxed-return attributes))
+                              (let ((args (make-gensym-list (length arg-types))))
+                                `(lambda ,args (funcall ',name ,@args)))))))
 
 (defstruct (fun-type-annotation
             (:copier nil))
@@ -179,13 +229,17 @@
           (labels ((annotation-p (x)
                      (typep x '(or (cons (member function function-designator modifying
                                           inhibit-flushing))
-                                (member type-specifier))))
+                                (member type-specifier proper-sequence proper-list
+                                 proper-or-dotted-list proper-or-circular-list))))
                    (strip-annotation (x)
                      (if (consp x)
                          (ecase (car x)
                            ((function function-designator) (car x))
                            ((modifying inhibit-flushing) (cadr x)))
-                         x))
+                         (case x
+                           (proper-sequence 'sequence)
+                           ((proper-list proper-or-dotted-list proper-or-circular-list) 'list)
+                           (t x))))
                    (process-positional (type)
                      (incf i)
                      (cond ((annotation-p type)
@@ -239,16 +293,20 @@
 
 ;;;; generic type inference methods
 
-(defun symeval-derive-type (node &aux (args (basic-combination-args node))
+(defun maybe-find-free-var (name)
+  (let ((found (gethash name (free-vars *ir1-namespace*))))
+    (unless (eq found :deprecated)
+      found)))
+
+(defun symbol-value-derive-type (node &aux (args (basic-combination-args node))
                                       (lvar (pop args)))
   (unless (and lvar (endp args))
-    (return-from symeval-derive-type))
+    (return-from symbol-value-derive-type))
   (if (constant-lvar-p lvar)
       (let* ((sym (lvar-value lvar))
              (var (maybe-find-free-var sym))
              (local-type (when var
-                           (let ((*lexenv* (node-lexenv node)))
-                             (lexenv-find var type-restrictions))))
+                           (lexenv-find var type-restrictions :lexenv (node-lexenv node))))
              (global-type (info :variable :type sym)))
         (if local-type
             (type-intersection local-type global-type)
@@ -266,14 +324,10 @@
   (let ((lvar (car (last (combination-args call)))))
     (when lvar (lvar-type lvar))))
 
-;;; Derive the result type according to the float contagion rules, but
-;;; always return a float. This is used for irrational functions that
-;;; preserve realness of their arguments.
-(defun result-type-float-contagion (call)
-  (declare (type combination call))
-  (reduce #'numeric-contagion (combination-args call)
-          :key #'lvar-type
-          :initial-value (specifier-type 'single-float)))
+(defun result-type-nth-arg (n)
+  (lambda (call)
+    (let ((lvar (nth n (combination-args call))))
+      (when lvar (lvar-type lvar)))))
 
 (defun simplify-list-type (type &key preserve-dimensions)
   ;; Preserve all the list types without dragging
@@ -294,16 +348,23 @@
 ;;; N'th argument. If arg is a list, result is a list. If arg is a
 ;;; vector, result is a vector with the same element type.
 (defun sequence-result-nth-arg (n &key preserve-dimensions
-                                       preserve-vector-type)
+                                       preserve-vector-type
+                                       string-designator)
   (lambda (call)
     (declare (type combination call))
-    (let ((lvar (nth (1- n) (combination-args call))))
+    (let ((lvar (nth n (combination-args call))))
       (when lvar
         (let ((type (lvar-type lvar)))
-          (cond ((simplify-list-type type
-                                     :preserve-dimensions preserve-dimensions))
+          (cond ((and (not string-designator)
+                      (simplify-list-type type
+                                          :preserve-dimensions preserve-dimensions)))
                 ((not (csubtypep type (specifier-type 'vector)))
-                 nil)
+                 (cond ((not string-designator) nil)
+                       ((csubtypep type (specifier-type 'character))
+                        (specifier-type `(simple-string 1)))
+                       ((and (constant-lvar-p lvar)
+                             (symbolp (lvar-value lvar)))
+                        (ctype-of (symbol-name (lvar-value lvar))))))
                 (preserve-vector-type
                  type)
                 (t
@@ -319,7 +380,7 @@
 (defun result-type-specifier-nth-arg (n)
   (lambda (call)
     (declare (type combination call))
-    (let ((lvar (nth (1- n) (combination-args call))))
+    (let ((lvar (nth n (combination-args call))))
       (when (and lvar (constant-lvar-p lvar))
         (careful-specifier-type (lvar-value lvar))))))
 
@@ -337,7 +398,7 @@
 (defun creation-result-type-specifier-nth-arg (n)
   (lambda (call)
     (declare (type combination call))
-    (let ((lvar (nth (1- n) (combination-args call))))
+    (let ((lvar (nth n (combination-args call))))
       (when (and lvar (constant-lvar-p lvar))
         (let* ((specifier (lvar-value lvar))
                (lspecifier (if (atom specifier) (list specifier) specifier)))
@@ -415,12 +476,11 @@
                                        (process-dim (array-type-dimensions type))))
                        (t (return '*))))
                    (process-dim (dim)
-                     (let ((length (car dim)))
-                       (if (and (singleton-p dim)
-                                (integerp length))
+                     (if (typep dim '(cons integer null))
+                         (let ((length (car dim)))
                            (setf max (max max length)
-                                 min (min min length))
-                           (return '*)))))
+                                 min (min min length)))
+                         (return '*))))
             ;; If type derivation were able to notice that non-simple arrays can
             ;; be mutated (changing the type), we could safely use LVAR-TYPE on
             ;; any vector type. But it doesn't notice.
@@ -430,44 +490,65 @@
             (max-dim (lvar-type lvar))
             (values max min))))))
 
-(defun position-derive-type (call)
-  (let ((dim (sequence-lvar-dimensions (second (combination-args call)))))
-    (when (integerp dim)
-      (specifier-type `(or (integer 0 (,dim)) null)))))
-
-(defun count-derive-type (call)
-  (let ((dim (sequence-lvar-dimensions (second (combination-args call)))))
-    (when (integerp dim)
-      (specifier-type `(integer 0 ,dim)))))
-
 ;;; This used to be done in DEFOPTIMIZER DERIVE-TYPE, but
 ;;; ASSERT-CALL-TYPE already asserts the ARRAY type, so it gets an extra
 ;;; assertion that may not get eliminated and requires extra work.
-(defun array-call-type-deriver (call trusted &optional set)
+(defun array-call-type-deriver (call trusted &optional set row-major-aref)
   (let* ((fun (combination-fun call))
          (type (lvar-fun-type fun))
          (policy (lexenv-policy (node-lexenv call)))
          (args (combination-args call)))
     (when (fun-type-p type)
-      (flet ((assert-type (arg type &optional set)
-               (when (if set
-                         (assert-modifying-lvar-type arg type
-                                                     (lvar-fun-name fun)
-                                                     policy)
-                         (assert-lvar-type arg type policy))
+      (flet ((assert-type (arg type &optional set index)
+               (when (cond (index
+                            (assert-array-index-lvar-type arg type policy))
+                           (t
+                            (when set
+                              (add-annotation arg
+                                              (make-lvar-modified-annotation :caller (lvar-fun-name fun))))
+                            (assert-lvar-type arg type policy)))
                  (unless trusted (reoptimize-lvar arg)))))
         (let ((required (fun-type-required type)))
           (when set
             (assert-type (pop args)
                          (pop required)))
           (assert-type (pop args)
-                       (type-intersection
-                        (pop required)
-                        (specifier-type `(array * ,(length args))))
+                       (if row-major-aref
+                           (pop required)
+                           (type-intersection
+                            (pop required)
+                            (let ((rank (length args)))
+                              (when (>= rank array-rank-limit)
+                                (setf (combination-kind call) :error)
+                                (compiler-warn "More subscripts for ~a (~a) than ~a (~a)"
+                                               (combination-fun-debug-name call)
+                                               rank
+                                               'array-rank-limit
+                                               array-rank-limit)
+                                (return-from array-call-type-deriver))
+                              (specifier-type `(array * ,rank)))))
                        set)
           (loop for type in required
-                do (assert-type (pop args) type))
+                do
+                (assert-type (pop args) type nil (or (not (and set row-major-aref))
+                                                     args)))
           (loop for type in (fun-type-optional type)
-                do (assert-type (pop args) type))
+                do (assert-type (pop args) type nil t))
           (loop for subscript in args
-                do (assert-type subscript (fun-type-rest type))))))))
+                do (assert-type subscript (fun-type-rest type) nil t)))))))
+
+(defun append-call-type-deriver (call trusted)
+  (let* ((policy (lexenv-policy (node-lexenv call)))
+         (args (combination-args call))
+         (list-type (specifier-type 'list)))
+    ;; All but the last argument should be proper lists
+    (loop for (arg next) on args
+          while next
+          do
+          (add-annotation
+           arg
+           (make-lvar-proper-sequence-annotation
+            :kind 'proper-list))
+          (when (and (assert-lvar-type arg list-type policy)
+                     (not trusted))
+            (reoptimize-lvar arg)))))
