@@ -817,11 +817,12 @@
   (defined-type *universal-type* :type ctype)
   ;; where the TYPE information came from (in order, from strongest to weakest):
   ;;  :DECLARED, from a declaration.
+  ;;  :DECLARED-VERIFY
   ;;  :DEFINED-HERE, from examination of the definition in the same file.
   ;;  :DEFINED, from examination of the definition elsewhere.
   ;;  :DEFINED-METHOD, implicit, piecemeal declarations from CLOS.
   ;;  :ASSUMED, from uses of the object.
-  (where-from :assumed :type (member :declared :assumed :defined-here :defined :defined-method))
+  (where-from :assumed :type (member :declared :declared-verify :assumed :defined-here :defined :defined-method))
   ;; list of the REF nodes for this leaf
   (refs () :type list)
   ;; For tracking whether to warn about unused variables:
@@ -829,7 +830,7 @@
   ;; SET if there was a set but no REF.
   ;; T if there was a REF.
   ;; This may be non-nil when REFS and SETS are null, since code can be deleted.
-  (ever-used nil :type (member nil set t))
+  (ever-used nil :type (member nil set t initial-unused))
   ;; True if declared dynamic-extent.
   (dynamic-extent nil :type boolean)
   ;; some kind of info used by the back end
