@@ -168,7 +168,8 @@
 
 (define-vop (make-closure)
   (:args (function :to :save :scs (descriptor-reg)))
-  (:info length stack-allocate-p)
+  (:info label length stack-allocate-p)
+  (:ignore label)
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:temporary (:sc non-descriptor-reg :offset nl3-offset) pa-flag)
   (:results (result :scs (descriptor-reg)))
@@ -212,12 +213,6 @@
   (:generator 1
     (inst li result unbound-marker-widetag)))
 
-(define-vop (make-funcallable-instance-tramp)
-  (:args)
-  (:results (result :scs (any-reg)))
-  (:generator 1
-    (inst addi result null-tn (make-fixup 'funcallable-instance-tramp :assembly-routine*))))
-
 (define-vop (fixed-alloc)
   (:args)
   (:info name words type lowtag stack-allocate-p)
