@@ -70,9 +70,10 @@
                     (t
                      (delete-1 tn prev setter))))))
              (used-p (tn)
-               (or (tn-reads tn) (tn-writes tn)
-                   (member (tn-kind tn) '(:component :environment))
-                   (not (zerop (sbit aliases (tn-number tn))))))
+               (and (neq (tn-kind tn) :unused)
+                    (or (tn-reads tn) (tn-writes tn)
+                        (member (tn-kind tn) '(:component :environment))
+                        (not (zerop (sbit aliases (tn-number tn)))))))
              (delete-1 (tn prev setter)
                (if prev
                    (setf (tn-next prev) (tn-next tn))
@@ -221,7 +222,8 @@
 ;;; EMIT-MOVES-AND-COERCIONS. That's wasteful.
 (defun make-constant-tn (constant &optional force-boxed)
   (declare (type constant constant))
-  (or (leaf-info constant)
+  (or (and (tn-p (leaf-info constant))
+           (leaf-info constant))
       (multiple-value-bind (immed null-offset)
           (immediate-constant-sc (constant-value constant))
         ;; currently NULL-OFFSET is used only on ARM64
@@ -231,8 +233,8 @@
                                       immed
                                       null-offset
                                       (specifier-type 'null)))))
-             (setf (tn-leaf tn) constant
-                   (leaf-info constant) tn))
+              (setf (tn-leaf tn) constant
+                    (leaf-info constant) tn))
             (let* ((boxed (or (not immed)
                               (boxed-immediate-sc-p immed)))
                    (component (component-info *component-being-compiled*))
@@ -251,10 +253,13 @@
               ;; because liveness depends on pointer tracing without looking at code-fixups.
               (when (and sc
                          (or (not immed)
+                             #+permgen (typep (constant-value constant) 'layout)
                              #+immobile-space
                              (let ((val (constant-value constant)))
                                (or (and (symbolp val) (not (sb-vm:static-symbol-p val)))
-                                   (typep val 'wrapper)))))
+                                   (typep val 'layout))))
+                         #+(or arm64 x86-64)
+                         (not (eql (constant-value constant) $0f0)))
                 (let ((constants (ir2-component-constants component)))
                   (setf (tn-offset res)
                         (vector-push-extend constant constants))))
@@ -341,30 +346,44 @@
 
 ;;;; TN referencing
 
+(defmacro link-tn-ref (write-p tn ref)
+  `(cond (,write-p
+          (let ((w (tn-writes ,tn)))
+            (when w
+              (setf (tn-ref-prev w) ,ref))
+            (setf (tn-ref-next ref) w
+                  (tn-writes tn) ,ref)))
+         (t
+          (let ((r (tn-reads ,tn)))
+            (when r
+              (setf (tn-ref-prev r) ,ref))
+            (setf (tn-ref-next ,ref) r
+                  (tn-reads tn) ,ref)))))
+
 ;;; Make a TN-REF that references TN and return it. WRITE-P should be
 ;;; true if this is a write reference, otherwise false. All we do
 ;;; other than calling the constructor is add the reference to the
 ;;; TN's references.
 (defun reference-tn (tn write-p)
   (declare (type tn tn) (type boolean write-p))
-  (let ((res (make-tn-ref tn write-p)))
+  (let ((ref (make-tn-ref tn write-p)))
     (unless (eql (tn-kind tn) :unused)
       (when (tn-primitive-type tn)
-        (aver (setf (tn-ref-type res) (tn-type tn))))
-      (if write-p
-          (push-in tn-ref-next res (tn-writes tn))
-          (push-in tn-ref-next res (tn-reads tn))))
-    res))
+        (aver (setf (tn-ref-type ref) (tn-type tn))))
+      (link-tn-ref write-p tn ref))
+    ref))
 
 (defun reference-tn-refs (refs write-p)
   (when refs
     (let* ((first (reference-tn (tn-ref-tn refs) write-p))
            (prev first))
+      (setf (tn-ref-type first) (tn-ref-type refs))
       (loop for tn-ref = (tn-ref-across refs) then (tn-ref-across tn-ref)
             while tn-ref
             do
-            (let ((ref (reference-tn  (tn-ref-tn tn-ref) write-p)))
-              (setf (tn-ref-across prev) ref)
+            (let ((ref (reference-tn (tn-ref-tn tn-ref) write-p)))
+              (setf (tn-ref-across prev) ref
+                    (tn-ref-type ref) (tn-ref-type tn-ref))
               (setq prev ref)))
       first)))
 
@@ -385,13 +404,39 @@
         first)
       more))
 
+;;; Copy the tn-ref-type of the TNs.
+(defun reference-tn-ref-list (tn-refs write-p &optional more)
+  (declare (list tn-refs) (type boolean write-p) (type (or tn-ref null) more))
+  (if tn-refs
+      (let* ((first (reference-tn (tn-ref-tn (first tn-refs)) write-p))
+             (prev first))
+        (setf (tn-ref-type first) (tn-ref-type (first tn-refs)))
+        (dolist (tn-ref (rest tn-refs))
+          (let ((res (reference-tn (tn-ref-tn tn-ref) write-p)))
+            (setf (tn-ref-across prev) res
+                  (tn-ref-type res) (tn-ref-type tn-ref))
+            (setq prev res)))
+        (setf (tn-ref-across prev) more)
+        first)
+      more))
+
 ;;; Remove Ref from the references for its associated TN.
 (defun delete-tn-ref (ref)
   (declare (type tn-ref ref))
-  (if (tn-ref-write-p ref)
-      (deletef-in tn-ref-next (tn-writes (tn-ref-tn ref)) ref)
-      (deletef-in tn-ref-next (tn-reads (tn-ref-tn ref)) ref))
-  (values))
+  (let ((tn (tn-ref-tn ref))
+        (prev (tn-ref-prev ref))
+        (next (tn-ref-next ref)))
+    (cond ((tn-ref-write-p ref)
+           (if prev
+               (setf (tn-ref-next prev) next)
+               (setf (tn-writes tn) next)))
+          (t
+           (if prev
+               (setf (tn-ref-next prev) next)
+               (setf (tn-reads tn) next))))
+    (when next
+      (setf (tn-ref-prev next) prev))
+    (setf (tn-ref-prev ref) nil)))
 
 ;;; Do stuff to change the TN referenced by Ref. We remove Ref from its
 ;;; old TN's refs, add ref to TN's refs, and set the TN-REF-TN.
@@ -399,10 +444,8 @@
   (declare (type tn-ref ref) (type tn tn))
   (delete-tn-ref ref)
   (setf (tn-ref-tn ref) tn)
-  (if (tn-ref-write-p ref)
-      (push-in tn-ref-next ref (tn-writes tn))
-      (push-in tn-ref-next ref (tn-reads tn)))
-  (values))
+  (link-tn-ref (tn-ref-write-p ref) tn ref)
+  nil)
 
 ;;;; miscellaneous utilities
 
