@@ -242,100 +242,167 @@
 (defun default-unknown-values (vop values nvals node rbx move-temp)
   (declare (type (or tn-ref null) values)
            (type unsigned-byte nvals))
-  (let ((type (sb-c::basic-combination-derived-type node)))
-    (cond
-      ((<= nvals 1)
-       (note-this-location vop :single-value-return)
+  (multiple-value-bind (type name leaf) (sb-c::lvar-fun-type (sb-c::basic-combination-fun node))
+   (let* ((verify (and leaf
+                        (policy node (and (>= safety 1)
+                                          (= debug 3)))
+                        (memq (sb-c::leaf-where-from leaf) '(:declared-verify :defined-here))))
+           (type (if verify
+                     (if (fun-type-p type)
+                         (fun-type-returns type)
+                         *wild-type*)
+                     (sb-c::node-derived-type node)))
+           (min-values (values-type-min-value-count type))
+           (max-values (values-type-max-value-count type))
+           (trust (or (and (= min-values 0)
+                           (= max-values call-arguments-limit))
+                      (not verify))))
+     (flet ((check-nargs ()
+              (assemble ()
+                (let* ((*location-context* (list* name
+                                                  (type-specifier type)
+                                                  (make-restart-location SKIP)))
+                       (err-lab (generate-error-code vop 'invalid-arg-count-error))
+                       (min (and (> min-values 0)
+                                 min-values))
+                       (max (and (< max-values call-arguments-limit)
+                                 max-values)))
+                  (cond ((not min)
+                         (if (zerop max)
+                             (inst test :dword rcx-tn rcx-tn)
+                             (inst cmp :dword rcx-tn (fixnumize max)))
+                         (inst jmp :ne err-lab))
+                        (max
+                         (if (zerop min)
+                             (setf move-temp rcx-tn)
+                             (inst lea :dword move-temp (ea (fixnumize (- min)) rcx-tn)))
+                         (inst cmp :dword move-temp (fixnumize (- max min)))
+                         (inst jmp :a err-lab))
+                        (t
+                         (cond ((= min 1)
+                                (inst test :dword rcx-tn rcx-tn)
+                                (inst jmp :e err-lab))
+                               ((plusp min)
+                                (inst cmp :dword rcx-tn (fixnumize min))
+                                (inst jmp :b err-lab))))))
+                SKIP)))
        (cond
-         ((<= (sb-kernel:values-type-max-value-count type)
-              register-arg-count))
-         ((not (sb-kernel:values-type-may-be-single-value-p type))
-          (inst mov rsp-tn rbx))
+         ((<= nvals 1)
+          (note-this-location vop :single-value-return)
+          (cond
+            ((and trust
+                  (<= (sb-kernel:values-type-max-value-count type)
+                      register-arg-count)))
+            ((and trust
+                  (not (sb-kernel:values-type-may-be-single-value-p type)))
+             (inst mov rsp-tn rbx))
+            (t
+             (inst cmov :c rsp-tn rbx)
+             (unless trust
+               (inst mov move-temp (fixnumize 1))
+               (inst cmov :nc rcx-tn move-temp)
+               (check-nargs)))))
+         ((<= nvals register-arg-count)
+          (note-this-location vop :unknown-return)
+          (when (or (not trust)
+                    (sb-kernel:values-type-may-be-single-value-p type))
+            (assemble ()
+              (inst jmp :c regs-defaulted)
+              ;; Default the unsupplied registers.
+              (let* ((2nd-tn-ref (tn-ref-across values))
+                     (2nd-tn (tn-ref-tn 2nd-tn-ref))
+                     (2nd-tn-live (neq (tn-kind 2nd-tn) :unused)))
+                (when 2nd-tn-live
+                  (inst mov 2nd-tn nil-value))
+                (when (> nvals 2)
+                  (loop
+                    for tn-ref = (tn-ref-across 2nd-tn-ref)
+                    then (tn-ref-across tn-ref)
+                    for count from 2 below register-arg-count
+                    unless (eq (tn-kind (tn-ref-tn tn-ref)) :unused)
+                    do
+                    (inst mov :dword (tn-ref-tn tn-ref)
+                          (if 2nd-tn-live 2nd-tn nil-value)))))
+              (inst mov rbx rsp-tn)
+              regs-defaulted))
+
+          (when (or (not trust)
+                    (< register-arg-count
+                       (sb-kernel:values-type-max-value-count type)))
+            (inst mov rsp-tn rbx))
+          (unless trust
+            (inst mov move-temp (fixnumize 1))
+            (inst cmov :nc rcx-tn move-temp)
+            (check-nargs)))
          (t
-          (inst cmov :c rsp-tn rbx))))
-      ((<= nvals register-arg-count)
-       (note-this-location vop :unknown-return)
-       (when (sb-kernel:values-type-may-be-single-value-p type)
-         (let ((regs-defaulted (gen-label)))
-           (inst jmp :c regs-defaulted)
-           ;; Default the unsupplied registers.
-           (let* ((2nd-tn-ref (tn-ref-across values))
-                  (2nd-tn (tn-ref-tn 2nd-tn-ref))
-                  (2nd-tn-live (neq (tn-kind 2nd-tn) :unused)))
-             (when 2nd-tn-live
-               (inst mov 2nd-tn nil-value))
-             (when (> nvals 2)
-               (loop
-                 for tn-ref = (tn-ref-across 2nd-tn-ref)
-                 then (tn-ref-across tn-ref)
-                 for count from 2 below register-arg-count
-                 unless (eq (tn-kind (tn-ref-tn tn-ref)) :unused)
-                 do
-                 (inst mov :dword (tn-ref-tn tn-ref)
-                       (if 2nd-tn-live 2nd-tn nil-value)))))
-           (inst mov rbx rsp-tn)
-           (emit-label regs-defaulted)))
-       (when (< register-arg-count
-                (sb-kernel:values-type-max-value-count type))
-         (inst mov rsp-tn rbx)))
-      (t
-       (collect ((defaults))
-         (let ((default-stack-slots (gen-label))
-               (used-registers
-                 (loop for i from 1 below register-arg-count
-                       for tn = (tn-ref-tn (setf values (tn-ref-across values)))
-                       unless (eq (tn-kind tn) :unused)
-                       collect tn
-                       finally (setf values (tn-ref-across values))))
-               (used-stack-slots-p
-                 (loop for ref = values then (tn-ref-across ref)
-                       while ref
-                       thereis (neq (tn-kind (tn-ref-tn ref)) :unused))))
-          (assemble ()
-            (note-this-location vop :unknown-return)
-            ;; If it returned exactly one value the registers and the
-            ;; stack slots need to be filled with NIL.
-            (cond (used-stack-slots-p
-                   (inst jmp :nc default-stack-slots))
-                  (t
-                   (inst jmp :c regs-defaulted)
-                   (loop for null = nil-value then (car used-registers)
-                         for reg in used-registers
-                         do (inst mov :dword reg null))
-                   (move rbx rsp-tn)
-                   (inst jmp defaulting-done)))
-            REGS-DEFAULTED
-            (do ((i register-arg-count (1+ i))
-                 (val values (tn-ref-across val)))
-                ((null val))
-              (let ((tn (tn-ref-tn val)))
-                (unless (eq (tn-kind tn) :unused)
-                  (let ((default-lab (gen-label)))
-                    (defaults (cons default-lab tn))
-                    ;; Note that the max number of values received
-                    ;; is assumed to fit in a :dword register.
-                    (inst cmp :dword rcx-tn (fixnumize i))
-                    (inst jmp :be default-lab)
-                    (sc-case tn
-                      (control-stack
-                       (loadw move-temp rbx (frame-word-offset (+ sp->fp-offset i)))
-                       (inst mov tn move-temp))
+          (collect ((defaults))
+            (let ((default-stack-slots (gen-label))
+                  (used-registers
+                    (loop for i from 1 below register-arg-count
+                          for tn = (tn-ref-tn (setf values (tn-ref-across values)))
+                          unless (eq (tn-kind tn) :unused)
+                          collect tn
+                          finally (setf values (tn-ref-across values))))
+                  (used-stack-slots-p
+                    (loop for ref = values then (tn-ref-across ref)
+                          while ref
+                          thereis (neq (tn-kind (tn-ref-tn ref)) :unused))))
+              (assemble ()
+                (note-this-location vop :unknown-return)
+                (unless trust
+                  (inst mov move-temp (fixnumize 1))
+                  (inst cmov :nc rcx-tn move-temp))
+                ;; If it returned exactly one value the registers and the
+                ;; stack slots need to be filled with NIL.
+                (cond ((and trust
+                            (> min-values 1)))
+                      (used-stack-slots-p
+                       (inst jmp :nc default-stack-slots))
                       (t
-                       (loadw tn rbx (frame-word-offset (+ sp->fp-offset i)))))))))
-            DEFAULTING-DONE
-            (move rsp-tn rbx)
-            (let ((defaults (defaults)))
-              (when defaults
-                (assemble (:elsewhere)
-                  (emit-label default-stack-slots)
-                  (loop for null = nil-value then (car used-registers)
-                        for reg in used-registers
-                        do (inst mov :dword reg null))
-                  (move rbx rsp-tn)
-                  (dolist (default defaults)
-                    (emit-label (car default))
-                    (inst mov (cdr default) nil-value))
-                  (inst jmp defaulting-done)))))))))))
+                       (inst jmp :c regs-defaulted)
+                       (loop for null = nil-value then (car used-registers)
+                             for reg in used-registers
+                             do (inst mov :dword reg null))
+                       (inst jmp done)))
+                REGS-DEFAULTED
+                (do ((i register-arg-count (1+ i))
+                     (val values (tn-ref-across val)))
+                    ((null val))
+                  (let ((tn (tn-ref-tn val)))
+                    (unless (eq (tn-kind tn) :unused)
+                      (when (or (not trust)
+                                (>= i min-values))
+                        (let ((default-lab (gen-label)))
+                          (defaults (cons default-lab tn))
+                          ;; Note that the max number of values received
+                          ;; is assumed to fit in a :dword register.
+                          (inst cmp :dword rcx-tn (fixnumize i))
+                          (inst jmp :be default-lab)))
+                      (sc-case tn
+                        (control-stack
+                         (loadw move-temp rbx (frame-word-offset (+ sp->fp-offset i)))
+                         (inst mov tn move-temp))
+                        (t
+                         (loadw tn rbx (frame-word-offset (+ sp->fp-offset i))))))))
+                DEFAULTING-DONE
+                (move rsp-tn rbx)
+                (unless trust
+                  (check-nargs))
+                DONE
+                (let ((defaults (defaults)))
+                  (when defaults
+                    (assemble (:elsewhere)
+                      (when (or (not trust)
+                                (<= min-values 1))
+                        (emit-label default-stack-slots)
+                        (loop for null = nil-value then (car used-registers)
+                              for reg in used-registers
+                              do (inst mov :dword reg null))
+                        (move rbx rsp-tn))
+                      (dolist (default defaults)
+                        (emit-label (car default))
+                        (inst mov (cdr default) nil-value))
+                      (inst jmp defaulting-done)))))))))))))
 
 ;;;; unknown values receiving
 
@@ -1403,41 +1470,6 @@
                    ((plusp min)
                     (inst cmp :dword nargs (fixnumize min))
                     (inst jmp :b err-lab))))))))
-
-;; Signal an error about an untagged number.
-;; These are pretty much boilerplate and could be generic except:
-;; - the names of the SCs could differ between backends (or maybe not?)
-;; - in the "/c" case, the older backends don't eval the errcode
-;; And the 6 vops above ought to be generic too...
-;; FIXME: there are still some occurrences of
-;;  note: doing signed word to integer coercion
-;; in regard to SB-C::%TYPE-CHECK-ERROR. Figure out why.
-(define-vop (type-check-error/word)
-  (:policy :fast-safe)
-  (:translate sb-c::%type-check-error)
-  (:args (object :scs (signed-reg unsigned-reg))
-         ;; Types are trees of symbols, so 'any-reg' is not
-         ;; really possible.
-         (type :scs (any-reg descriptor-reg constant)))
-  (:arg-types untagged-num * (:constant t))
-  (:info *location-context*)
-  (:vop-var vop)
-  (:save-p :compute-only)
-  ;; cost is a smidgen less than type-check-error
-  ;; otherwise this does not get selected.
-  (:generator 999
-    (error-call vop 'object-not-type-error object type)))
-(define-vop (type-check-error/word/c)
-  (:policy :fast-safe)
-  (:translate sb-c::%type-check-error/c)
-  (:args (object :scs (signed-reg unsigned-reg)))
-  (:arg-types untagged-num (:constant symbol) (:constant t))
-  (:info errcode *location-context*)
-  (:vop-var vop)
-  (:save-p :compute-only)
-  (:generator 899 ; smidgen less than type-check-error/c
-    (error-call vop errcode object)))
-
 ;;; Single-stepping
 
 (defun emit-single-step-test ()
