@@ -38,43 +38,44 @@
   (:info dest not-p flags)
   (:vop-var vop)
   (:generator 0
-    (when (cdr flags)
-      ;; Specifying multiple flags is an extremely confusing convention that supports
-      ;; floating-point inequality tests utilizing the flag register in an unusual way,
-      ;; as documented in the COMISS and COMISD instructions:
-      ;; "the ZF, PF, and CF flags in the EFLAGS register according
-      ;;  to the result (unordered, greater than, less than, or equal)"
-      ;; I think it would have been better if, instead of allowing more than one flag,
-      ;; we passed in a pseudo-condition code such as ':sf<=' and deferred to this vop
-      ;; to interpret the abstract condition in terms of how to CPU sets the bits
-      ;; in those particular instructions.
-      ;; Note that other architectures allow only 1 flag in branch-if, if it works at all.
-      ;; Enable this assertion if you need to sanity-check the preceding claim.
-      ;; There's really no "dynamic" reason for it to fail.
-      #+nil
-      (aver (memq (vop-name (sb-c::vop-prev vop))
-                  '(<single-float <double-float <=single-float <=double-float
-                    >single-float >double-float >=single-float >=double-float
-                    =/single-float =/double-float))))
-     (when (eq (car flags) 'not)
-       (pop flags)
-       (setf not-p (not not-p)))
-     (cond ((null (rest flags))
-              (inst jmp
-                    (if not-p
-                        (negate-condition (first flags))
-                        (first flags))
-                    dest))
-           (not-p
-              (let ((not-lab (gen-label))
-                    (last    (car (last flags))))
-                (dolist (flag (butlast flags))
-                  (inst jmp flag not-lab))
-                (inst jmp (negate-condition last) dest)
-                (emit-label not-lab)))
-           (t
-              (dolist (flag flags)
-                (inst jmp flag dest))))))
+    (let ((flags (conditional-flags-flags flags)))
+      (when (cdr flags)
+        ;; Specifying multiple flags is an extremely confusing convention that supports
+        ;; floating-point inequality tests utilizing the flag register in an unusual way,
+        ;; as documented in the COMISS and COMISD instructions:
+        ;; "the ZF, PF, and CF flags in the EFLAGS register according
+        ;;  to the result (unordered, greater than, less than, or equal)"
+        ;; I think it would have been better if, instead of allowing more than one flag,
+        ;; we passed in a pseudo-condition code such as ':sf<=' and deferred to this vop
+        ;; to interpret the abstract condition in terms of how to CPU sets the bits
+        ;; in those particular instructions.
+        ;; Note that other architectures allow only 1 flag in branch-if, if it works at all.
+        ;; Enable this assertion if you need to sanity-check the preceding claim.
+        ;; There's really no "dynamic" reason for it to fail.
+        #+nil
+        (aver (memq (vop-name (sb-c::vop-prev vop))
+                    '(<single-float <double-float <=single-float <=double-float
+                      >single-float >double-float >=single-float >=double-float
+                      =/single-float =/double-float))))
+      (when (eq (car flags) 'not)
+        (pop flags)
+        (setf not-p (not not-p)))
+      (cond ((null (rest flags))
+             (inst jmp
+                   (if not-p
+                       (negate-condition (first flags))
+                       (first flags))
+                   dest))
+            (not-p
+             (let ((not-lab (gen-label))
+                   (last    (car (last flags))))
+               (dolist (flag (butlast flags))
+                 (inst jmp flag not-lab))
+               (inst jmp (negate-condition last) dest)
+               (emit-label not-lab)))
+            (t
+             (dolist (flag flags)
+               (inst jmp flag dest)))))))
 
 (define-vop (multiway-branch-if-eq)
   ;; TODO: also accept signed-reg, unsigned-reg, character-reg
@@ -83,7 +84,7 @@
   (:info labels otherwise key-type keys test-vop-name)
   (:temporary (:sc unsigned-reg) table)
   (:temporary (:sc unsigned-reg) temp)
-  (:args-var x-tn-ref)
+  (:arg-refs x-tn-ref)
   (:generator 10
     (let* ((key-derived-type (tn-ref-type x-tn-ref))
            (ea)
@@ -165,27 +166,37 @@
   (:info flags)
   (:temporary (:sc unsigned-reg) temp)
   (:generator 0
-    (let ((not-p (eq (first flags) 'not)))
+    (let* ((flags (conditional-flags-flags flags))
+           (not-p (eq (first flags) 'not))
+           (size))
       (when not-p (pop flags))
       (when (location= res then)
         (rotatef then else)
         (setf not-p (not not-p)))
       (flet ((load-immediate (dst constant-tn
                               &optional (sc-reg dst))
-               ;; Can't use ZEROIZE, since XOR will affect the flags.
-               (inst mov dst
-                     (encode-value-if-immediate constant-tn
-                                                (sc-is sc-reg any-reg descriptor-reg)))))
+               (let ((encode (encode-value-if-immediate constant-tn
+                                                        (sc-is sc-reg any-reg descriptor-reg))))
+                 (if (typep encode '(unsigned-byte 31))
+                     (unless size
+                       (setf size :dword))
+                     (setf size :qword))
+                 ;; Can't use ZEROIZE, since XOR will affect the flags.
+                 (inst mov dst encode))))
         (cond ((null (rest flags))
-               (if (sc-is else immediate)
-                   (load-immediate res else)
-                   (move res else))
-               (when (sc-is then immediate)
-                 (load-immediate temp then res)
-                 (setf then temp))
-               (inst cmov (if not-p
-                              (negate-condition (first flags))
-                              (first flags))
+               (cond ((sc-is else immediate)
+                      (load-immediate res else))
+                     (t
+                      (setf size :qword)
+                      (move res else)))
+               (cond ((sc-is then immediate)
+                      (load-immediate temp then res)
+                      (setf then temp))
+                     (t
+                      (setf size :qword)))
+               (inst cmov size (if not-p
+                                   (negate-condition (first flags))
+                                   (first flags))
                      res
                      then))
               (not-p
@@ -195,8 +206,9 @@
                         (setf else temp))
                       (load-immediate res then))
                      ((location= else res)
-                      (inst xchg else then)
-                      (rotatef else then))
+                      (move temp else)
+                      (move res then)
+                      (setf else temp))
                      (t
                       (move res then)))
                (when (sc-is else immediate)
@@ -315,12 +327,12 @@
 ;;; Note: a constant-tn is allowed in CMP; it uses an EA displacement,
 ;;; not immediate data.
 (define-vop (if-eq)
-  (:args (x :scs (any-reg descriptor-reg control-stack immediate))
+  (:args (x :scs (any-reg descriptor-reg control-stack))
          (y :scs (any-reg descriptor-reg control-stack immediate constant)))
   (:conditional :e)
   (:policy :fast-safe)
   (:translate eq)
-  (:args-var x-tn-ref)
+  (:arg-refs x-tn-ref)
   (:temporary (:sc unsigned-reg) temp)
   (:generator 6
     (cond
@@ -407,10 +419,10 @@
   (assert (eql other-pointer-lowtag #b1111))
   ;; This is also assumed in src/runtime/x86-64-assem.S
   (assert (eql (min bignum-widetag ratio-widetag single-float-widetag double-float-widetag
-                    complex-widetag complex-single-float-widetag complex-double-float-widetag)
+                    complex-rational-widetag complex-single-float-widetag complex-double-float-widetag)
                bignum-widetag))
   (assert (eql (max bignum-widetag ratio-widetag single-float-widetag double-float-widetag
-                    complex-widetag complex-single-float-widetag complex-double-float-widetag)
+                    complex-rational-widetag complex-single-float-widetag complex-double-float-widetag)
                complex-double-float-widetag)))
 
 ;;; Most uses of EQL are transformed into a non-generic form, but when we need
@@ -448,10 +460,8 @@
 
     ;; If the widetags are not the same, return false.
     ;; Using a :dword compare gets us the bignum length check almost for free
-    ;; unless the length's representation requires more than 3 bytes.
-    ;; It sounds like a :qword compare would be the right thing, but remember
-    ;; one header bit acts as a concurrent GC mark bit in all headered objects,
-    ;; though we're not really using it yet. (We are, but not concurrently)
+    ;; unless the length's representation requires more 4 bytes.
+    ;; I bet nobody would mind if MAXIMUM-BIGNUM-LENGTH were #xFFFFFF.
     (inst mov :dword rax (ea (- other-pointer-lowtag) x))
     (inst cmp :dword rax (ea (- other-pointer-lowtag) y))
     (inst jmp :ne done) ; negative
