@@ -82,10 +82,10 @@
 (defconstant msan-temp-reg-number 0)
 
 ;;; The encoding anomaly for r12 makes it a perfect choice for the card table base.
-;;; It will never be used with a constant displacement.
+;;; It will seldom be used with a constant displacement.
 (define-symbol-macro card-table-reg 12)
 (define-symbol-macro gc-card-table-reg-tn r12-tn)
-(define-symbol-macro card-index-mask (make-fixup nil :gc-barrier))
+(define-symbol-macro card-index-mask (make-fixup nil :card-table-index-mask))
 
 (macrolet ((defreg (name offset size)
              (declare (ignore size))
@@ -187,9 +187,7 @@
 ;;; Start from 2, for the old RBP (aka OCFP) and return address
 (define-storage-base stack :unbounded :size 2 :size-increment 1)
 (define-storage-base constant :non-packed)
-(define-storage-base immediate-constant :non-packed)
-(define-storage-base noise :unbounded :size 2)
-)
+(define-storage-base immediate-constant :non-packed))
 
 ;;;; SC definitions
 
@@ -245,12 +243,6 @@
   #+sb-simd-pack-256
   (single-avx2-stack stack :element-size 4)
 
-  ;;
-  ;; magic SCs
-  ;;
-
-  (ignore-me noise)
-
   ;;
   ;; things that can go in the integer registers
   ;;
@@ -464,11 +456,6 @@
             (symbol-value (symbolicate register-arg-name "-TN")))
           *register-arg-names*))
 
-#-sb-xc-host
-(defun pseudostatic-immobile-symbol-p (symbol)
-  (and (immobile-space-obj-p symbol)
-       (= (generation-of symbol) +pseudo-static-generation+)))
-
 ;;; If value can be represented as an immediate constant, then return
 ;;; the appropriate SC number, otherwise return NIL.
 (defun immediate-constant-sc (value)
@@ -477,23 +464,19 @@
          character)
      immediate-sc-number)
     (symbol ; Symbols in static and immobile space are immediate
-     (when (or ;; With #+immobile-symbols, all symbols are in immobile-space.
-               ;; And the cross-compiler always uses immobile-space if enabled.
-               #+(or immobile-symbols (and immobile-space sb-xc-host)) t
-
-               ;; Otherwise, if #-immobile-symbols, and the symbol is pseudostatic
-               ;; in immobile space, it is an immediate value.
-               ;; If compiling to memory, the symbol's address alone suffices.
-               #+(and (not sb-xc-host) immobile-space (not immobile-symbols))
-               (or (pseudostatic-immobile-symbol-p value)
-                   (locally (declare (notinline sb-c::producing-fasl-file))
-                     (and (not (sb-c::producing-fasl-file))
-                          (immobile-space-obj-p value))))
-
-               (static-symbol-p value))
+     (when (or (static-symbol-p value)
+               ;; The cross-compiler always uses immobile-space if it exists.
+               #+(and immobile-space sb-xc-host) t
+               ;; With #+immobile-symbols, all interned symbols are in immobile-space.
+               #+immobile-symbols (sb-xc:symbol-package value)
+               #-sb-xc-host
+               (if (immobile-space-obj-p value)
+                   (or (= (generation-of value) +pseudo-static-generation+)
+                       ;; If compiling to memory, the symbol's address alone suffices.
+                       (locally (declare (notinline sb-c::producing-fasl-file))
+                         (not (sb-c::producing-fasl-file))))))
        immediate-sc-number))
-    #+metaspace (sb-vm:layout (bug "Can't reference layout as a constant"))
-    #+(and immobile-space (not metaspace)) (wrapper immediate-sc-number)
+    #+compact-instance-header (layout immediate-sc-number)
     (single-float
        (if (eql value $0f0) fp-single-zero-sc-number fp-single-immediate-sc-number))
     (double-float
@@ -535,8 +518,8 @@
           (symbol   (if (static-symbol-p val)
                         (+ nil-value (static-symbol-offset val))
                         (make-fixup val :immobile-symbol)))
-          #+(and immobile-space (not metaspace))
-          (wrapper
+          #+(or immobile-space permgen)
+          (layout
            (make-fixup val :layout))
           (character (if tag
                          (logior (ash (char-code val) n-widetag-bits)
@@ -549,7 +532,7 @@
                  bits)))
           (structure-object
            (if (eq val sb-lockless:+tail+)
-               (progn (aver tag) sb-vm::lockfree-list-tail-value)
+               (progn (aver tag) (+ static-space-start lockfree-list-tail-value-offset))
                (bug "immediate structure-object ~S" val)))))
       tn))
 
@@ -627,16 +610,36 @@
        (if (or
             (valid-funtype '((mod 64) word) '*)
             (valid-funtype '((mod 64) signed-word) '*))
-           (values :transform '(lambda (index integer) (%logbitp index integer)))
+           (values :direct nil)
            (values :default nil)))
+      (truncate
+       (destructuring-bind (n &optional d) (sb-c::basic-combination-args node)
+         (if (and d
+                  (constant-lvar-p d)
+                  (power-of-two-p (lvar-value d))
+                  (and (csubtypep (sb-c::lvar-type n) (specifier-type 'signed-word))
+                       (not (csubtypep (sb-c::lvar-type n) (specifier-type 'word)))))
+             (values :maybe nil)
+             (values :default nil))))
+      (%dpb
+       (flet ((validp (type result-type)
+                (valid-funtype `((constant-arg integer)
+                                 (constant-arg (integer 1 1))
+                                 ,type
+                                 ,type)
+                               result-type)))
+         (if (and (or (validp 'signed-word 'signed-word)
+                      (validp 'word 'word))
+                  (not (destructuring-bind (new size posn integer) (sb-c::basic-combination-args node)
+                         (declare (ignore new size integer))
+                         (constant-lvar-p posn))))
+             (values :direct nil)
+             (values :default nil))))
       (t
        (values :default nil)))))
 
 (defvar *register-names* +qword-register-names+)
 
-(defmacro unbound-marker-bits ()
-  (logior (+ sb-vm:static-space-start #x100) unbound-marker-widetag))
-
 ;;; See WRITE-FUNINSTANCE-PROLOGUE in x86-64-vm.
 ;;; There are 4 bytes available in the imm32 operand of a dummy MOV instruction.
 ;;; (It's a valid instruction on the theory that illegal opcodes might cause
@@ -644,3 +647,19 @@
 (defmacro compact-fsc-instance-hash (fin)
   `(sap-ref-32 (int-sap (get-lisp-obj-address ,fin))
                (+ (ash 3 word-shift) 4 (- fun-pointer-lowtag))))
+
+(eval-when (:compile-toplevel)
+  (let ((locs (sb-c::sc-locations (gethash 'any-reg sb-c:*backend-sc-names*))))
+    ;; These locations are not saved by WITH-REGISTERS-PRESERVED
+    ;; because Lisp can't treat them as general purpose.
+    ;; By design they are also (i.e. must be) nonvolatile aross C call.
+    (aver (not (logbitp 12 locs)))
+    #-gs-seg (aver (not (logbitp 13 locs)))))
+
+#+sb-xc-host
+(setq *backend-cross-foldable-predicates*
+      '(power-of-two-p))
+
+#+nil
+(define-cond-sc 32-bit-immediate immediate
+  (typep (tn-value tn) '(signed-byte 32)))
