@@ -170,7 +170,7 @@
 (export 'compute-cache-index) ; for a test
 (defun compute-cache-index (cache layouts)
   (macrolet ((fetch-hash (x)
-               `(let ((hash (wrapper-clos-hash ,x)))
+               `(let ((hash (layout-clos-hash ,x)))
                   (when (zerop hash) (return-from compute-cache-index nil))
                   hash)))
     (let ((index (fetch-hash (car layouts))))
@@ -181,6 +181,7 @@
 
 ;;; The hash values are specifically designed so that LOGAND of all hash
 ;;; inputs is zero if and only if at least one hash is invalid.
+(eval-when (:compile-toplevel :load-toplevel :execute)
 (defun cache-mixer-expression (operator operands associativep)
   (named-let recurse ((operands operands))
     (let ((n (length operands)))
@@ -197,7 +198,7 @@
                            ,(recurse (subseq operands half-n)))))
             (t
              (recurse (cons `(,operator ,(first operands) ,(second operands))
-                            (cddr operands))))))))
+                            (cddr operands)))))))))
 
 ;;; Emit code that does lookup in cache bound to CACHE-VAR using
 ;;; layouts bound to LAYOUT-VARS. Go to MISS-TAG on event of a miss or
@@ -207,18 +208,17 @@
 ;;; In other words, produces inlined code for COMPUTE-CACHE-INDEX when
 ;;; number of keys and presence of values in the cache is known
 ;;; beforehand.
-(defglobal *hash-vars*
-  (make-array 6 :initial-contents
-              (loop for i from 0 below 6 collect (make-symbol (format nil "H~D" i)))))
+(eval-when (:compile-toplevel :load-toplevel :execute)
 (defun emit-cache-lookup (cache-var layout-vars miss-tag value-var)
-  (declare (muffle-conditions code-deletion-note))
   (multiple-value-bind (probe n-vector n-depth n-mask
                       MATCH-WRAPPERS EXIT-WITH-HIT)
       (values '#:probe '#:vect '#:depth '#:mask '#:try '#:hit)
     (let* ((num-keys (length layout-vars))
+           ;; just to avoid always freshly gensym'ing everything
+           (premade-hash-vars #(#:H0 #:H1 #:H2 #:H3 #:H4 #:H5))
            (hash-vars (loop for i from 0 below num-keys
-                            collect (if (< i #.(length *hash-vars*))
-                                        (aref *hash-vars* i)
+                            collect (if (< i (length premade-hash-vars))
+                                        (aref premade-hash-vars i)
                                         (make-symbol (format nil "H~D" i)))))
            (pointer
             ;; We don't need POINTER if the cache has 1 key and no value,
@@ -228,7 +228,7 @@
             (when (or (> num-keys 1) value-var) (make-symbol "PTR")))
            (line-size (power-of-two-ceiling (+ num-keys (if value-var 1 0)))))
       ;; Why not use PROG* ? are we expressly trying to avoid a new block?
-      `(let* (,@(mapcar (lambda (x y) `(,x (wrapper-clos-hash ,y))) hash-vars layout-vars)
+      `(let* (,@(mapcar (lambda (x y) `(,x (layout-clos-hash ,y))) hash-vars layout-vars)
               (,n-mask (cache-mask ,cache-var))
               (,probe (if (zerop ,(cache-mixer-expression 'logand hash-vars t))
                           (go ,miss-tag)
@@ -264,7 +264,7 @@
             (setf ,probe (next-cache-index ,n-mask ,probe ,line-size))
             ,@(if pointer `((setf ,pointer ,probe)))
             (go ,MATCH-WRAPPERS)
-            ,EXIT-WITH-HIT)))))
+            ,EXIT-WITH-HIT))))))
 
 ;;; Probes CACHE for LAYOUTS.
 ;;;
@@ -295,7 +295,7 @@
       (case key-count
        (1
         (let* ((layout (if (%instancep key) key (car key)))
-               (hash (wrapper-clos-hash layout))
+               (hash (layout-clos-hash layout))
                (index (logand hash mask)))
           (unless (= hash 0)
             (probe-loop (eq (svref vector (truly-the index index)) layout)))))
@@ -376,6 +376,36 @@
         ;; Make a smaller one, then
         (make-cache :key-count key-count :value value :size (ceiling size 2)))))
 
+(defmacro dotimes-fixnum ((var count &optional (result nil)) &body body)
+  `(dotimes (,var (the fixnum ,count) ,result)
+     (declare (fixnum ,var))
+     ,@body))
+
+(define-load-time-global *pcl-misc-random-state* (make-random-state))
+
+(declaim (inline random-fixnum))
+(defun random-fixnum ()
+  (random (1+ most-positive-fixnum)
+          (load-time-value *pcl-misc-random-state*)))
+
+;;; Lambda which executes its body (or not) randomly. Used to drop
+;;; random cache entries.
+;;; This formerly punted with slightly greater than 50% probability,
+;;; and there was a periodicity to the nonrandomess.
+;;; If that was intentional, it should have been commented to that effect.
+(defmacro randomly-punting-lambda (lambda-list &body body)
+  (with-unique-names (drops drop-pos)
+    `(let ((,drops (random-fixnum)) ; means a POSITIVE fixnum
+           (,drop-pos sb-vm:n-positive-fixnum-bits))
+       (declare (fixnum ,drops)
+                (type (mod #.sb-vm:n-fixnum-bits) ,drop-pos))
+       (lambda ,lambda-list
+         (when (logbitp (the unsigned-byte (decf ,drop-pos)) ,drops)
+           (locally ,@body))
+         (when (zerop ,drop-pos)
+           (setf ,drops (random-fixnum)
+                 ,drop-pos sb-vm:n-positive-fixnum-bits))))))
+
 ;;;; Copies and expands the cache, dropping any invalidated or
 ;;;; incomplete lines.
 (defun copy-and-expand-cache (cache layouts value)
@@ -387,10 +417,7 @@
       (setf length (* 2 length)))
     (tagbody
      :again
-       ;; Blow way the old vector first, so a GC potentially triggered by
-       ;; MAKE-ARRAY can collect it.
-       (setf (cache-vector copy) #()
-             (cache-vector copy) (make-cache-storage length)
+       (setf (cache-vector copy) (make-cache-storage length)
              (cache-depth copy) 0
              (cache-mask copy) (compute-cache-mask length (cache-line-size cache))
              (cache-limit copy) (compute-limit (/ length (cache-line-size cache))))
@@ -437,14 +464,14 @@
       ;; Check if the line is in use, and check validity of the keys.
       (let ((key1 (svref vector index)))
         (when (cache-key-p key1)
-          (if (zerop (wrapper-clos-hash key1))
+          (if (zerop (layout-clos-hash key1))
               ;; First key invalid.
               (return-from cache-has-invalid-entries-p t)
               ;; Line is in use and the first key is valid: check the rest.
               (loop for offset from 1 below key-count
                     do (let ((thing (svref vector (+ index offset))))
                          (when (or (not (cache-key-p thing))
-                                   (zerop (wrapper-clos-hash thing)))
+                                   (zerop (layout-clos-hash thing)))
                            ;; Incomplete line or invalid layout.
                            (return-from cache-has-invalid-entries-p t)))))))
       ;; Line empty of valid, onwards.
@@ -552,6 +579,30 @@
                  :mask mask
                  :limit (cache-limit cache))))
 
+(defun %struct-typecase-miss (object cache-cell)
+  ;; If the layout of OBJECT is already in the cache, then returned the value
+  ;; from the cache. Otherwise try to compute and memoize the index.
+  (when (%instancep object)
+    (let ((layout (%instance-layout object))
+          (cache (car cache-cell))
+          (clause-index 0))
+      (macrolet ((lookup () (emit-cache-lookup 'cache '(layout) 'miss 'clause-index)))
+        (tagbody
+           (lookup)
+           (return-from %struct-typecase-miss clause-index)
+         MISS))
+      (setq clause-index 1)
+      (dovector (layouts (cdr cache-cell))
+        (dolist (test-layout layouts)
+          (when (sb-c::%structure-is-a layout test-layout) ; success, to try memoize it
+            (let ((key (list layout))) ; insert this object's layout
+              (unless (try-update-cache cache key clause-index)
+                (setf (car cache-cell)
+                      (copy-and-expand-cache cache key clause-index))))
+            (return-from %struct-typecase-miss clause-index)))
+        (incf clause-index))))
+  0)
+
 ;;;; For debugging & collecting statistics.
 
 (defun map-all-caches (function)
@@ -567,7 +618,7 @@
           (when (fboundp name)
             (let ((fun (fdefinition name)))
               (when (typep fun 'generic-function)
-                (let ((cache (gf-dfun-cache fun)))
+                (let ((cache (funcall 'gf-dfun-cache fun)))
                   (when cache
                     (funcall function name cache)))))))))))
 
@@ -588,7 +639,7 @@
          (total-lines (/ size line-size))
          (total-n-keys 0) ; a "key" is a tuple of layouts
          (n-dirty 0) ; lines that have an unbound marker but are not wholly empty
-         (n-obsolete 0) ; lines that need to be evicted due to 0 in a wrapper-clos-hash
+         (n-obsolete 0) ; lines that need to be evicted due to 0 in a layout-clos-hash
          (histogram
            (when compute-histogram
              (make-array (1+ (cache-limit cache)) :initial-element 0))))
@@ -601,7 +652,7 @@
                         (n-misses 0))
                    (cond ((find-if-not #'cache-key-p layouts)
                           (incf n-dirty))
-                         ((find 0 layouts :key #'wrapper-clos-hash)
+                         ((find 0 layouts :key #'layout-clos-hash)
                           (incf n-obsolete))
                          (t
                           (incf total-n-keys)
@@ -626,7 +677,7 @@
                      (multiple-value-bind (count capacity n-dirty n-obsolete histogram)
                          (cache-statistics cache t)
                        (list histogram count capacity n-dirty n-obsolete cache)))
-                   (sb-vm::list-allocated-objects :all :test #'cache-p)))
+                   (sb-vm:list-allocated-objects :all :test #'cache-p)))
          (n 0)
          (n-shown 0)
          (histogram-column-width
