@@ -49,14 +49,9 @@ static void* get_free_page() {
     --free_page;
     if (free_page < next_free_page)
         lose("Needed more space to GC");
-    page_table[free_page].type = PAGE_TYPE_UNBOXED;
+    set_page_type(page_table[free_page], PAGE_TYPE_UNBOXED);
     char* mem = page_address(free_page);
-#ifdef LISP_FEATURE_DARWIN_JIT
-    // Might need to alter MMU-based protection
-    zeroize_pages_if_needed(free_page, free_page, 0);
-#elif defined LISP_FEATURE_WIN32
-    os_commit_memory(page_address(free_page), GENCGC_PAGE_BYTES);
-#endif
+    prepare_pages(1, free_page, free_page, PAGE_TYPE_UNBOXED, -1);
     return mem;
 }
 
@@ -289,6 +284,7 @@ void execute_full_mark_phase()
 #endif
     trace_object((lispobj*)NIL_SYMBOL_SLOTS_START);
     scav_static_range((lispobj*)STATIC_SPACE_OBJECTS_START, static_space_free_pointer);
+    scav_static_range((lispobj*)PERMGEN_SPACE_START, permgen_space_free_pointer);
 #ifndef LISP_FEATURE_IMMOBILE_SPACE
     // if NO immobile-space, then text space is equivalent to static space
     scav_static_range((lispobj*)TEXT_SPACE_START, text_space_highwatermark);
@@ -456,7 +452,7 @@ void dispose_markbits() {
         gc_assert((page_table[page].type & PAGE_TYPE_MASK) == PAGE_TYPE_UNBOXED);
         gc_assert(!page_bytes_used(page));
         set_page_need_to_zero(page, 1);
-        page_table[page].type = FREE_PAGE_FLAG;
+        set_page_type(page_table[page], FREE_PAGE_FLAG);
       }
 }
 
